//
// File:        CommandLineDriver.java
// Package:     gov.llnl.babel
// Revision:    $Revision: 4836 $
// Modified:    $Date: 2005-08-20 22:30:15 -0700 (Sat, 20 Aug 2005) $
// Description: main driver for the babel IDL compiler
//
// Copyright (c) 2000-2004, The Regents of the University of Calfornia.
// Produced at the Lawrence Livermore National Laboratory.
// Written by the Components Team <components@llnl.gov>
// UCRL-CODE-2002-054
// All rights reserved.
// 
// This file is part of Babel. For more information, see
// http://www.llnl.gov/CASC/components/. Please read the COPYRIGHT file
// for Our Notice and the LICENSE file for the GNU Lesser General Public
// License.
// 
// This program is free software; you can redistribute it and/or modify it
// under the terms of the GNU Lesser General Public License (as published by
// the Free Software Foundation) version 2.1 dated February 1999.
// 
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the IMPLIED WARRANTY OF
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the terms and
// conditions of the GNU Lesser General Public License for more details.
// 
// You should have recieved a copy of the GNU Lesser General Public License
// along with this program; if not, write to the Free Software Foundation,
// Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

package gov.llnl.babel;

import gov.llnl.babel.ast.SIDLFile;
import gov.llnl.babel.backend.BuildGenerator;
import gov.llnl.babel.backend.CodeGenerationException;
import gov.llnl.babel.backend.CodeGenerationFactory;
import gov.llnl.babel.backend.CodeGenerator;
import gov.llnl.babel.backend.FileListener;
import gov.llnl.babel.backend.FileManager;
import gov.llnl.babel.cli.CommandLineDictionary;
import gov.llnl.babel.cli.CommandLineExtension;
import gov.llnl.babel.cli.InvalidOptionException;
import gov.llnl.babel.cli.NameCollisionException;
import gov.llnl.babel.msg.ErrorFormatter;
import gov.llnl.babel.msg.MsgList;
import gov.llnl.babel.parsers.sidl.Parser;
import gov.llnl.babel.parsers.sidl.SIDLException;
import gov.llnl.babel.parsers.sidl2.ParseTree2ASTVisitor;
import gov.llnl.babel.parsers.sidl2.SIDLDumpVisitor;
import gov.llnl.babel.parsers.sidl2.SIDLParser;
import gov.llnl.babel.parsers.sidl2.SIDLTouchUpVisitor;
import gov.llnl.babel.parsers.sidl2.ParseTreeNode;
import gov.llnl.babel.repository.RepositoryException;
import gov.llnl.babel.repository.RepositoryFactory;
import gov.llnl.babel.symbols.AssertionException;
import gov.llnl.babel.symbols.RegexMatch;
import gov.llnl.babel.symbols.Symbol;
import gov.llnl.babel.symbols.SymbolID;
import gov.llnl.babel.symbols.SymbolNotFoundException;
import gov.llnl.babel.symbols.SymbolRedefinitionException;
import gov.llnl.babel.symbols.SymbolTable;
import gov.llnl.babel.url.URLUtilities;
import gov.llnl.babel.visitor.CollisionChecker;
import gov.llnl.babel.visitor.DumpVisitor;

import java.io.BufferedInputStream;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;
import java.util.StringTokenizer;

/**
 * This class is the main driver for running Babel from the command line.
 */
public class CommandLineDriver {
 /**
  * The command line parser.
  */
  private static UserOptions d_option_parser = UserOptions.getInstance();

  /**
   * The file generation manager.
   */
  private FileManager d_file_manager = null;

  /**
   * The symbol table
   */
  private SymbolTable d_symbol_table = null;

  /**
   * The handle to the singleton instance of this class.
   */
  private static CommandLineDriver s_my_instance = null;

  /**
   * 
   */
  private static MsgList d_msgs = null;

  /**
   * The protected singleton constructor instantiates the options parser,
   * <code>UserOptions</code>.
   * 
   * @see UserOptions
   */
  protected CommandLineDriver() {
    d_msgs = new MsgList();
  }
  
  /**
    * Return the singleton instance of this class.
    */ 
  private static CommandLineDriver getInstance() {
    if (s_my_instance == null) {
      s_my_instance = new CommandLineDriver();
    }
    return s_my_instance;
  }

  /**
   * Extract the options from the command line arguments and perform associated
   * set up as appropriate.
   * 
   * @see BabelConfiguration
   */
  private int processCommandline(String args[]) {
    SwitchDictionary dict = new SwitchDictionary();
    try {
      d_option_parser.registerCommandLineSwitches(dict);
    } catch(NameCollisionException nce) {
      System.err.
        println
        ("Babel: Internal Babel command line switches have a name space conflict!");
      return 1;
    }
    loadExtensions(dict);
    String[]remainingArgs;
    try {
      remainingArgs = dict.executeCommandLineSwitches(args);
    }
    catch(InvalidOptionException ioe) {
      return 1;
    }
    if (dict.getHelpPrinted() || d_option_parser.getVersionPrinted()) {
      return 0;
    }

    if ((d_option_parser.getNumRequired() == 0)
        || (remainingArgs.length == 0)) {
      dict.printHelpText();
      return 1;
    }

    if (BabelConfiguration.getInstance().generateStdlib()) {
      if (BabelConfiguration.getInstance().generateClient()
          && "python".equals(BabelConfiguration.getInstance()
                             .getTargetLanguage())) {
        System.err.
          println
          ("Babel: Warning: Allowing concurrent request for python client and SIDL standard library generation.");
      }
      else {
        if (!(BabelConfiguration.getInstance().generateServer()
              && "c".equals(BabelConfiguration.getInstance().
                            getTargetLanguage()))) {
          System.err.
            println
            ("Babel: Error: --generate-sidl-stdlib only makes sense with --server=c or --client=python.");
          return 1;
        }
      }
    }

    /*
     * Now continue processing the remainder of the arguments and create
     * associated processing elements as needed.
     */
    resolveRepositoryPaths();
    try {
      resolveRemainingArgs(remainingArgs, 0);
    }
    catch(CollectionException ex) {
      System.err.println("Babel: FATAL: Unable to resolve remaining args.");
      ex.printStackTrace(System.err);
      return 1;
    }

    /*
     * Stop now if only checking parsing.
     */
    if (BabelConfiguration.getInstance().parseCheckOnly()) {
      return 0;
    }

    /*
     * Generate the source code.
     */
    generateSourceCode();

    return 0;
  }

  /**
   * Create <code>Repository</code> instances based on the user-specified
   * repository path and add them to the <code>SymbolTable</code> singleton.
   * 
   * @see SymbolTable
   */
  private void resolveRepositoryPaths() {
    String repoPath = BabelConfiguration.getInstance().getRepositoryPath();
    RepositoryFactory the_factory = RepositoryFactory.getInstance();

    d_symbol_table = SymbolTable.getInstance();

    for (StringTokenizer tokenizer = new StringTokenizer(repoPath, ";");
         tokenizer.hasMoreTokens();) {
      String path = tokenizer.nextToken();
      if ((path == null) || (path.equals("")) || (path.equals("null"))) {
        continue;
      }
      try {
        d_symbol_table.addSymbolResolver(the_factory.createRepository(path));
      }
      catch(RepositoryException ex) {
        System.err.
          println("Babel: Warning: Unable to create repository using " +
                  "path \"" + path + "\".  Will use default.");
        ex.printStackTrace(System.err);
      }
    }
  }

  /**
   * Initialize the file generation manager with configuration-specific
   * information as needed.
   * 
   * @see BabelConfiguration
   * @see FileManager
   */
  private void setupFileManager() {
    d_file_manager = FileManager.getInstance();
    try {
      d_file_manager.setFileGenerationRootDirectory(BabelConfiguration.
                                                    getInstance().
                                                    getOutputDirectory());
    } catch(CodeGenerationException ex) {
      System.err.
        println("Babel: Warning: Unable to set output directory. Will " +
                "use default.");
      ex.printStackTrace(System.err);
    }

    d_file_manager.setJavaStylePackageGeneration(BabelConfiguration.
                                                 getInstance().
                                                 makePackageSubdirs());
    d_file_manager.setGlueSubdirGeneration(BabelConfiguration.getInstance()
                                           .makeGlueSubdirs());

    try {
      d_file_manager.setVPathDirectory(BabelConfiguration.getInstance()
                                       .getVPathDirectory());
    } catch(CodeGenerationException ex) {
      System.err.println("Babel: Warning: Unable to set VPATH directory.");
      System.err.println(ex.getMessage());
      ex.printStackTrace(System.err);
    }
  }

  private CollectionException addToCollection(CollectionException ce,
                                              String uri, Exception ex) {
    if (ce == null) {
      ce = new CollectionException(uri, ex);
    }
    else {
      ce.addException(uri, ex);
    }
    return ce;
  }

  /**
   * Process the remainder of the command line, assuming the contents reflect
   * one or more SIDL files.
   */
  private void resolveRemainingArgs(String args[], int firstUnusedArg)
  throws CollectionException {
    /*
     * First parse the remaining arguments.
     */
    CollectionException ce = null;
    for (int i = firstUnusedArg; i < args.length; i++) {
      try {
        File f = new File(args[i]);
        if (f.exists() && f.isFile() && f.canRead()) {
          /*
           * Parse the URL for the readable file.
           */
          String filename = f.toURL().toString();
          if (!filename.endsWith(".sidl") && !filename.endsWith(".idl")) {
            System.out.
              println("Babel: Warning: Will probably not be able to " +
                      "parse \"" + filename +
                      "\" since only support parsing " + "SIDL files.");
          }
          parseURL(filename);
          BabelConfiguration.getInstance().setSIDLFileInput(true);
        }
        else if (f.exists() && f.isDirectory()) {
          /*
           * If it's a directory, assume user meant a symbol, so resolve
           * it.
           */
          resolveSymbolByString(args[i]);
        }
        else if (args[i].indexOf(':') != -1) {
          /*
           * Try it as a URL if there's a colon.
           */
          parseURL(URLUtilities.expandURL(args[i]));
          BabelConfiguration.getInstance().setSIDLFileInput(true);
        }
        else if (resolveSymbolByString(args[i]) == null) {
          /*
           * Otherwise, it must be tried as a symbol.
           */
          throw new CodeGenerationException("Babel: Error: \"" + args[i]
                                            +
                                            "\" fails to resolve as a symbol or file.");
        }
      }
      catch(Exception ex) {
        System.err.println(ex.getMessage());
        ex.printStackTrace();
        ce = addToCollection(ce, args[i], ex);
      }
    }                           // end for loop

    /*
     * Now resolve all the symbols by first handling those by SIDL base class.
     */
    if (d_symbol_table == null) {
      System.err.
        println("Babel: Warning: Symbol table was not populated prior " +
                "to symbol resolution. Attempting to obtain instance now.");
      d_symbol_table = SymbolTable.getInstance();
    }

    try {
      d_symbol_table.resolveSymbol("sidl");
      d_symbol_table.resolveAllReferences();
    }
    catch(Exception ex) {
      System.err.println(ex.getMessage());
      ex.printStackTrace();
      ce = addToCollection(ce, null, ex);
    }
    if (ce != null) {
      throw ce;
    }
  }

  /**
   * Open and parse contents of the specified URL.
   */
  private static void parseURL(String url) throws SIDLException,
    MalformedURLException, IOException
  {
    URL u = new URL(url);
    InputStream s = null;
    try {
      s = u.openStream();
      if (BabelConfiguration.getInstance().experimentalParser()) {
        SIDLParser parser = new SIDLParser(s);
        ParseTreeNode parseTree = parser.Start();
        parseTree.name = url;
        parseTree.jjtAccept(new SIDLTouchUpVisitor(System.out), null);
        parseTree.jjtAccept(new SIDLDumpVisitor(System.out), null);
        SIDLFile AST =
          (SIDLFile) parseTree.jjtAccept(new ParseTree2ASTVisitor(d_msgs), null);
        AST.accept(new DumpVisitor(System.out), null);
        AST.accept(new CollisionChecker(d_msgs), null);
        System.out.flush();
        d_msgs.print(System.err);
        if (d_msgs.fatal_message()) {
          System.exit(88);
        } else { 
          AST.accept(new DumpVisitor(System.out), null);
        }
      }
      else {
        Parser parser = new Parser(new BufferedInputStream(s));
        System.out.println("Babel: Parsing URL \"" + url + "\".");
        parser.setSourceURL(url);
        String[]warnings = parser.beginParse();
        if (warnings != null) {
          for (int w = 0; w < warnings.length; w++) {
            System.err.println("Babel: Warning: " + warnings[w]);
          }
        }
      }
    } catch(gov.llnl.babel.parsers.sidl2.ParseException e) {
      ErrorFormatter.formatParseError(e, url);
    }
    finally {
      if (s != null)
        s.close();
    }
  }

  /**
   * Try to parse string as a symbol. Return null if it fails.
   */
  private Symbol resolveSymbolByString(String symbol_string)
  throws NumberFormatException, SymbolRedefinitionException {
    int split = symbol_string.indexOf("-v");
    String symbol_name = (split > 0
                          && split <
                          symbol_string.length())? symbol_string.
      substring(0, split)
      : symbol_string;
    String version_name = (split > 0
                           && (split + 2) <
                           symbol_string.length())? symbol_string.
      substring(split + 2)
      : null;
    Symbol symbol = null;

    if ((version_name == null) || (version_name == "")) {
      symbol = d_symbol_table.resolveSymbol(symbol_name);
    }
    else {
      SymbolID id = new SymbolID(symbol_name,
                                 new gov.llnl.babel.symbols.
                                 Version(version_name));
      symbol = d_symbol_table.resolveSymbol(id);
    }

    if (symbol != null) {
      System.out.println("Babel: Resolved symbol \"" + symbol_string + "\".");
      symbol.setUserSpecified(true);

      if (symbol.isPackage()) {
        /*
         * Add patterns for all symbols specifed on the command line, so
         * that code will be generated for relevant subpackages and the
         * package contents (if the symbol is a package).
         */
        BabelConfiguration.getInstance().addIncluded(symbol_string + ".");
      }
    }

    return symbol;
  }

  /**
   * Generate the appropriate source code based on user configuration
   * parameters.
   */
  private void generateSourceCode() {
    if (d_symbol_table == null) {
      System.out.
        println("Babel: Warning: Symbol table was not populated prior " +
                "to generating source code. Obtaining instance now.");
      d_symbol_table = SymbolTable.getInstance();
    }
    Set symbols = d_symbol_table.getSymbolNames();

    if (BabelConfiguration.getInstance().excludeExternal()) {
      symbols = excludeExternalSymbols(symbols);
    }
    symbols = excludeSymbols(symbols);

    /*
     * Now that the symbols are loaded, it's time to make sure they are valid
     * _before_ any code can be generated.
     */
    try {
      d_symbol_table.resolveAllParents();
      d_symbol_table.validateAssertions();
    }
    catch(AssertionException aex) {
      System.err.
        println("Babel: FATAL: Unable to validate the assertions of " +
                "the symbols in the symbol table.");
      aex.printStackTrace(System.err);
      System.exit(1);
    }
    catch(SymbolNotFoundException snfe) {
      System.err.println("Babel: Warning: Cannot find all parent packages.");
      System.err.println(snfe.getMessage());
    }

    /*
     * NOW it should be safe to generate code.
     */
    List configs = BabelConfiguration.getConfigurations();
    Iterator j = configs.iterator();
    while (j.hasNext()) {
      BabelConfiguration.setCurrentConfiguration((BabelConfiguration) j.
                                                 next());
      setupFileManager();
      String targetLanguage = BabelConfiguration.getInstance()
        .getTargetLanguage();
      boolean generateText = BabelConfiguration.getInstance().generateText();
      Collection buildGens = getBuildGen(targetLanguage, !generateText);
      Collection listeners = new LinkedList();
      Iterator i = buildGens.iterator();
      try {
        while (i.hasNext()) {
          BuildGenerator mg = (BuildGenerator) i.next();
          if (mg instanceof FileListener) {
            listeners.add((FileListener) mg);
            d_file_manager.addListener((FileListener) mg);
          }
        }
        if (BabelConfiguration.getInstance().generateText()) {
          generateCode(symbols, "text", targetLanguage);
        }
        else {
          if (BabelConfiguration.getInstance().generateClient()) {
            generateClientSource(symbols);
          }
          else if (BabelConfiguration.getInstance().generateServer()) {
            generateServerSource(symbols);
          }
        }
        generateMakefiles(buildGens);
      }
      finally {
        // remove build generators before going another run
        d_file_manager.removeListeners(listeners);
      }
    }
  }

  private void generateCode(Set symbols, String type, String language) {
    // TLD-To check contents of symbols...printSymbolNames(symbols);
    try {
      CodeGenerator gen = getCodeGen(language, type);
      if (gen != null) {
        if (gen.getUserSymbolsOnly()) {
          symbols = filterSymbols(symbols);
        }
        gen.generateCode(symbols);
      }
      else {
        System.err.println("Babel: Warning: Unable to generate the " + type
                           + " sources in the " + language + " language.");
      }
    }
    catch(CodeGenerationException ex) {
      System.err.println("Babel: FATAL: Unable to generate " + type
                         + " code.");
      System.err.println(ex.getMessage());
      ex.printStackTrace(System.err);
      System.exit(1);
    }
  }

  /**
   * Generate the server source code.
   */
  private void generateServerSource(Set symbols) {
    /*
     * First generate the client code.
     */
    generateClientSource(symbols);

    /*
     * Now generate the server-specific code.
     */
    String targetLanguage = BabelConfiguration.getInstance()
      .getTargetLanguage();
    if (!(BabelConfiguration.getInstance().getSuppressIOR()
          || "ior".equals(targetLanguage))) {
      generateCode(symbols, "skel", "ior");
    }
    generateCode(symbols, "skel", targetLanguage);
  }

  /**
   * Generate the client source code.
   */
  private void generateClientSource(Set symbols) {
    String lang = BabelConfiguration.getInstance().getTargetLanguage();
    if (!(BabelConfiguration.getInstance().getSuppressIOR()
          || "ior".equals(lang))) {
      generateCode(symbols, "stub", "ior");
    }
    generateCode(symbols, "stub", lang);
  }

  private Collection getBuildGen(String myLang, boolean warning) {
    Collection buildGens = CodeGenerationFactory.getInstance()
      .getBuildGenerators(myLang);
    if (buildGens.size() == 0 && warning) {
      System.err.
        println("Babel: Warning: No build generator returned for the " +
                myLang + " language.");
    }
    return buildGens;
  }

  /**
   * Instantiate the appropriate code generator associated with the desired
   * language and mode.
   * 
   * @see CodeGenerator
   */
  private CodeGenerator getCodeGen(String myLang, String mode)
  throws CodeGenerationException {
    CodeGenerator cg;

      cg = CodeGenerationFactory.getInstance().getCodeGenerator(myLang, mode);
    if (cg == null) {
      System.err.
        println("Babel: Warning: No code generator returned for the " +
                myLang + "language and " + mode + " mode.");
    }
    else {
      cg.setName(myLang);
    }
    return cg;
  }

  public static boolean notExcluded(SymbolID id) {
    Iterator i = BabelConfiguration.getInstance().getExcludedList()
      .iterator();
    while (i.hasNext()) {
      if (((RegexMatch) i.next()).match(id))
        return false;
    }
    return true;
  }

  private static Set excludeSymbols(Collection symbols) {
    Set result = new HashSet(symbols.size());
    Iterator i = symbols.iterator();
    while (i.hasNext()) {
      SymbolID id = (SymbolID) i.next();
      if (notExcluded(id)) {
        result.add(id);
      }
    }
    return result;
  }

  public static boolean isIncluded(SymbolID id) {
    final String fullname = id.getFullName();
    Iterator incl = BabelConfiguration.getInstance().getIncludedList()
      .iterator();
    while (incl.hasNext()) {
      if (fullname.startsWith((String) incl.next()))
        return true;
    }
    return false;
  }

  private Set excludeExternalSymbols(Collection symbols) {
    Set result = new HashSet(symbols.size());
    if (BabelConfiguration.getInstance().excludeExternal()) {
      Iterator i = symbols.iterator();
      while (i.hasNext()) {
        SymbolID id = (SymbolID) i.next();
        if (d_symbol_table.lookupSymbol(id).getUserSpecified()) {
          result.add(id);
        }
        else {
          /*
           * Check if current symbol is contained within the scope of a
           * user-specified (on the command line) symbol.
           */
          if (isIncluded(id)) {
            d_symbol_table.lookupSymbol(id).setUserSpecified(true);
            result.add(id);
          }
        }
      }
    }
    else {
      result.addAll(symbols);
    }
    return result;
  }

  private static boolean filterSymbol(SymbolID id) {
    return !BabelConfiguration.isSIDLBaseClass(id);
  }

  /**
   * Depending on the configuration settings, return either all symbols or just
   * the user symbols, those symbols outside the SIDL namespace.
   */
  private static Set filterSymbols(Set symbols) {
    Set result;
    if (BabelConfiguration.getInstance().generateStdlib()) {
      result = symbols;
    }
    else {
      result = new HashSet(symbols.size());
      Iterator i = symbols.iterator();
      while (i.hasNext()) {
        SymbolID id = (SymbolID) i.next();
        if (filterSymbol(id))
          result.add(id);
      }
    }
    return result;
  }

  /**
   * Generate the appropriate makefile fragments.
   */
  private void generateMakefiles(Collection buildGens) {
    /*
     * Now generate the fragments.
     */
    Iterator i = buildGens.iterator();
    while (i.hasNext()) {
      BuildGenerator mg = (BuildGenerator) i.next();
      try {
        mg.createAll();
      }
      catch(IOException ex) {
        System.err.println("Babel: Warning: Unable to generate makefiles.");
        System.err.println(ex.getMessage());
        ex.printStackTrace(System.err);

      }
    }
  }

  /**
   * Print all symbols to output. This is obviously intended only to facilitate
   * debugging.
   */
  public void printSymbolNames(Set symbols) {
    System.out.println("Symbol set:");
    Iterator i = symbols.iterator();
    while (i.hasNext()) {
      SymbolID id = (SymbolID) i.next();
      System.out.println("  " + id.getFullName());
    }
  }

  /**
   * Load and create an instance of each class in listed in the comma separated
   * list stored in the Java property <code>babel.extensions</code>.
   */
  public static void loadExtensions(CommandLineDictionary dict) {
    CodeGenerationFactory cgf = CodeGenerationFactory.getInstance();
    StringTokenizer tok =
      new StringTokenizer(System.getProperty("babel.extensions", ""), ",");

    while (tok.hasMoreTokens()) {
      final String className = tok.nextToken();
      try {
        boolean matches = false;
        final Class cls = Class.forName(className);
        final Object obj = cls.newInstance();
        if (obj instanceof CodeGenerator) {
          cgf.registerCodeGenerator((CodeGenerator) obj);
          matches = true;
        }
        if (obj instanceof BuildGenerator) {
          cgf.registerBuildGenerator((BuildGenerator) obj);
          matches = true;
        }
        if (obj instanceof CommandLineExtension) {
          ((CommandLineExtension) obj).registerCommandLineSwitches(dict);
          matches = true;
        }
        if (!matches) {
          System.err.println("Babel: Warning: Extension " + className
                             + " has no recognized extensions to register.");
        }
      }
      catch(NameCollisionException nce) {
        System.err.
          println("Babel: Warning: Problem with command line extension " +
                  className + ".");
        System.err.println("Babel: Warning: " + nce.getMessage());
      }
      catch(ClassNotFoundException cnfe) {
        System.err.println("Babel: Warning: Unable to load extension "
                           + className + ".");
      }
      catch(LinkageError le) {
        System.err.println("Babel: Warning: Unable to load extension "
                           + className +
                           " due to dependencies on other classes.");
      }
      catch(IllegalAccessException iae) {
        System.err.
          println
          ("Babel: Warning: Unable to create an instance of extension " +
           className + ".");
      }
      catch(InstantiationException ie) {
        System.err.
          println("Babel: Warning: Cannot create an instance of extension " +
                  className +
                  " because it's an interface or abstract class.");
      }
      catch(SecurityException se) {
        System.err.println("Babel: Warning: Extension " + className
                           + " violates the current Java security policy.");
      }
    }
  }

  /****************************************************************************
   * Main babel entry point.
   */
  public static void main(String args[]) {
    int returnValue = 1;
    try {
      /*
       * Set up the configuration parameters based on command line arguments.
       */
      CommandLineDriver myDriver = getInstance();
      returnValue = myDriver.processCommandline(args);
    } catch(Throwable t) {
      System.err.println("Babel.CommandLineDriver: Unable to generate code.");
      String msg = t.getMessage();
      if (msg != null) {
        System.err.println(t.getMessage());
      }
      t.printStackTrace();
      Runtime.getRuntime().exit(1);
    }
    finally {
      /* workaround for JNI bug in Linux 1.3.1 JVM */
      Runtime.getRuntime().exit(returnValue);
    }
  }
}
