package gov.llnl.babel.ast;

import gov.llnl.babel.parsers.sidl2.SIDLParserTreeConstants;
import gov.llnl.babel.parsers.sidl2.ParseTreeNode;
import gov.llnl.babel.visitor.Visitor;

/**
 * General encapsulation of built-in attributes (such as <code>final</code> or
 * <code>abstract</code>) as well as user-defined attributes, including both
 * singletons, or key-value pairs.
 * <p>
 * 
 * Built-in varieties are assumed to have an implicit key with a leading
 * underscore for instance <code>nonblocking</code>, <code>oneway</code>,
 * <code>local</code> are all mutually exclusive values on the key
 * <code>_communication</code>. Use <code>getVisibleKey</code> if you
 * prefer not to be bothered with this implementation detail.
 * 
 * User defined varieties such as <code>%attrib{blue}</code> are assigned to
 * keys with the value of <code>null</code>. Note that keys tend to be
 * mutually exclusive... so <code>%attrib{color,color=blue}</code> will cause
 * a compilation error.
 * <p>
 * 
 * @author kumfert
 * 
 */
public class Attribute extends ASTNode implements Comparable {

  String d_key;

  String d_value;

  public Attribute(ParseTreeNode src) {
    this(src, null);
  }

  public Attribute(ParseTreeNode src, ASTNode parent) {
    super(src, parent);

    d_key = "_unknown";
    if (src.getID() == SIDLParserTreeConstants.JJTCUSTOMATTR) {
      int index = src.name.indexOf('=');
      if (index == -1) {
        d_value = src.name;
      } else {
        d_key = src.name.substring(0, index);
        d_value = src.name.substring(index + 1);
      }
    } else {
      d_value = src.name;
    }

    if (d_key.equals("_unknown")) {
      if (d_value.equals("final") || d_value.equals("static")
          || d_value.equals("abstract")) {
        d_key = "_dispatch";
      } else if (d_value.equals("local") || d_value.equals("nonblocking")
          || d_value.equals("oneway")) {
        d_key = "_communication";
      } else if (d_value.equals("copy")) {
        d_key = "_copy";
      } else {
        d_key = d_value;
        d_value = null;
      }
    }
  }

  public void setParent(ASTNode parent) {
    d_parent = parent;
  }

  public String getKey() {
    return d_key;
  }

  public boolean isBuiltin() {
    return d_key.startsWith("_");
  }

  public String getVisibleKey() {
    if (isBuiltin()) {
      return d_value;
    } else {
      return d_key;
    }
  }

  public String getValue() {
    return d_value;
  }

  public int compareTo(Object o) {
    return d_key.compareTo(((Attribute) o).d_key);
  }

  public String toString() {
    if (isBuiltin()) {
      return d_value;
    } else {
      if (d_value == null) {
        return d_key;
      } else {
        return d_key + "=" + d_value;
      }
    }
  }

  public Object accept(Visitor v, Object data) {
    // return v.visitArgument( this, data );
    return null;
  }
}
