//
// File: RMIStubSource.java
// Package: gov.llnl.babel.backend.rmi
// Revision: @(#) $Id: RMIStubSource.java 5146 2005-12-17 04:48:40Z leek2 $
// Description: generate IOR stub source to a pretty writer stream
//
// Copyright (c) 2000-2001, The Regents of the University of Calfornia.
// Produced at the Lawrence Livermore National Laboratory.
// Written by the Components Team <components@llnl.gov>
// UCRL-CODE-2002-054
// All rights reserved.
// 
// This file is part of Babel. For more information, see
// http://www.llnl.gov/CASC/components/. Please read the COPYRIGHT file
// for Our Notice and the LICENSE file for the GNU Lesser General Public
// License.
// 
// This program is free software; you can redistribute it and/or modify it
// under the terms of the GNU Lesser General Public License (as published by
// the Free Software Foundation) version 2.1 dated February 1999.
// 
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the IMPLIED WARRANTY OF
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the terms and
// conditions of the GNU Lesser General Public License for more details.
// 
// You should have recieved a copy of the GNU Lesser General Public License
// along with this program; if not, write to the Free Software Foundation,
// Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

package gov.llnl.babel.backend.rmi;

import gov.llnl.babel.backend.writers.LanguageWriter;
import gov.llnl.babel.backend.writers.LanguageWriterForC;
import gov.llnl.babel.backend.CodeGenerationException;
import gov.llnl.babel.backend.IOR;
import gov.llnl.babel.backend.c.C;
import gov.llnl.babel.backend.Utilities;
import gov.llnl.babel.symbols.Argument;
import gov.llnl.babel.symbols.Class;
import gov.llnl.babel.symbols.Interface;
import gov.llnl.babel.symbols.Extendable;
import gov.llnl.babel.symbols.Method;
import gov.llnl.babel.symbols.Symbol;
import gov.llnl.babel.symbols.SymbolID;
import gov.llnl.babel.symbols.SymbolTable;
import gov.llnl.babel.symbols.Type;
import gov.llnl.babel.BabelConfiguration;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

/**
 * Class <code>RMIStubSource</code> generates an IOR implementation source
 * file to a language writer output stream. The constructor takes a language
 * writer stream and method <code>generateCode</code> generates intermediate
 * object representation for the specified symbol to the output stream. The
 * language writer output stream is not closed by this object.
 */
public class RMIStubSource {
  private static int s_longestBuiltin;

  private static final String s_deleteBuiltin = IOR.getBuiltinName(IOR.DELETE);

  private static final String s_castBuiltin = IOR.getBuiltinName(IOR.CAST);

  private static final String s_execBuiltin = IOR.getBuiltinName(IOR.EXEC);

  private static final String s_getURLBuiltin = IOR.getBuiltinName(IOR.GETURL);

  private static final String s_raddRefBuiltin = IOR
      .getBuiltinName(IOR.RADDREF);

  private static final String s_isRemoteBuiltin = IOR
      .getBuiltinName(IOR.ISREMOTE);

  /**
   * Store the SymbolID for sidl.BaseClass, if the extendable being printed is
   * not abstract (i.e., is a concrete class).
   */
  SymbolID d_baseClass = null;

  /**
   * Store the SymbolID for sidl.ClassInfo, if the extendable being printed is
   * not abstract (i.e., is a concrete class).
   */
  SymbolID d_classInfo = null;

  /**
   * Store the SymbolID for sidl.ClassInfoI, if the extendable being printed is
   * not abstract (i.e., is a concrete class).
   */
  SymbolID d_classInfoI = null;

  static {
    s_longestBuiltin = 0;
    for (int j = 0; j < IOR.CLASS_BUILTIN_METHODS; ++j) {
      String mname = IOR.getBuiltinName(j);
      if (mname.length() > s_longestBuiltin) {
        s_longestBuiltin = mname.length();
      }
    }
  }

  private LanguageWriterForC d_writer;

  /**
   * This is a convenience utility function that writes the symbol source
   * information into the provided language writer output stream. The output
   * stream is not closed on exit. A code generation exception is thrown if an
   * error is detected.
   */
  public static void generateCode(Symbol symbol, LanguageWriterForC writer)
      throws CodeGenerationException {
    RMIStubSource source = new RMIStubSource(writer);
    source.generateCode(symbol);
  }

  /** Feature allows one to generate includes and code seperately.
   *  Currently this is a special feature for Python to avoid conflicts
   *  between the function connectI's and rmicast's in the sidl C binding,
   *  and the macros defined in Python.
   */
  public static void generateIncludes(Symbol symbol, LanguageWriterForC writer)
    throws CodeGenerationException {
    RMIStubSource source = new RMIStubSource(writer);
    if (symbol != null) {
      switch (symbol.getSymbolType()) {
      case Symbol.PACKAGE:
      case Symbol.ENUM:
        break;
      case Symbol.INTERFACE:
        // Interface version take Interface and Anonymous Class
        source.generateIncludes(((Interface) symbol).generateAnonymousClass());
        
        break;
      case Symbol.CLASS:
        source.generateIncludes((Extendable) symbol);
        
        break;
      }
    }
  }

  /** Feature allows one to generate includes and code seperately.
   *  Currently this is a special feature for Python to avoid conflicts
   *  between the function connectI's and rmicast's in the sidl C binding,
   *  and the macros defined in Python.
   */
  public static void generateCodeNoIncludes(Symbol symbol, LanguageWriterForC writer)
    throws CodeGenerationException {
    RMIStubSource source = new RMIStubSource(writer);
    if (symbol != null) {
      switch (symbol.getSymbolType()) {
      case Symbol.PACKAGE:
      case Symbol.ENUM:
        break;
      case Symbol.INTERFACE:
        // Interface version take Interface and Anonymous Class
        source.generateStaticVariables(((Interface) symbol).generateAnonymousClass());
        source.generateSource((Interface) symbol, ((Interface) symbol)
                       .generateAnonymousClass(), false);
        
        break;
      case Symbol.CLASS:
        source.generateStaticVariables((Extendable) symbol);
        source.generateSource((Extendable) symbol,false);
        break;
      }
    }
  }

  /**
   * Create a <code>RMIStubSource</code> object that will write symbol
   * information to the provided output writer stream.
   */
  public RMIStubSource(LanguageWriterForC writer) {
    d_writer = writer;
  }

  /**
   * Write IOR source information for the provided symbol to the language writer
   * output stream provided in the constructor. This method does not close the
   * language writer output stream and may be called for more than one symbol
   * (although the generated source may not be valid input for the C compiler).
   * A code generation exception is generated if an error is detected. No code
   * is generated for enumerated and package symbols.
   */
  public void generateCode(Symbol symbol) throws CodeGenerationException {
    if (symbol != null) {
      switch (symbol.getSymbolType()) {
      case Symbol.PACKAGE:
      case Symbol.ENUM:
        break;
      case Symbol.INTERFACE:
        // Interface version take Interface and Anonymous Class
        generateSource((Interface) symbol, ((Interface) symbol)
                       .generateAnonymousClass(),true);
        break;
      case Symbol.CLASS:
        generateSource((Extendable) symbol,true);
        break;
      }
    }
  }

  /**
   * Lookup the SymbolIDs for sidl.BaseClass, sidl.ClassInfo and
   * sidl.ClassInfoI.
   */
  private void lookupSymbolIDs() {
    SymbolTable table = SymbolTable.getInstance();
    d_baseClass = table.lookupSymbol(BabelConfiguration.getBaseClass())
        .getSymbolID();
    d_classInfo = table.lookupSymbol(BabelConfiguration.getClassInfo())
        .getSymbolID();
    d_classInfoI = table.lookupSymbol(BabelConfiguration.getClassInfoI())
        .getSymbolID();
  }

  /**
   * Generate the RStub source for a sidl symbol . The source file begins with a
   * banner and include files followed by a declaration of static methods and
   * (for a class) external methods expected in the skeleton file. For classes,
   * the source file then defines a number of functions (cast, delete,
   * initialize EPVs, new, init, and fini).
   */
  private void generateSource(Extendable ext, boolean includes) throws CodeGenerationException {
    /*
     * Generate the file banner and include files.
     */
    SymbolID id = ext.getSymbolID();
    Class cls = (Class) ext;
    SymbolTable table = SymbolTable.getInstance();

    if(includes) {
      generateIncludes(ext);
      generateStaticVariables(ext);
    }
    /*
     * Generate internal static variables and external references to be supplied
     * by the skeleton file.
     */
    

    /*
     * Generate remote method support for interfaces and classes.
     */
    //generateRemoteCastFunction(cls, null);
    IOR.generateCastFunction(cls, "self", d_writer, true,true);
    generateRemoteDeleteFunction(ext);
    generateRemoteGetURLFunction(ext);
    generateRemoteAddRefFunction(ext);
    generateIsLocalIsRemote(cls);
    generateRemoteExecFunction(ext);
    generateRemoteMethodBodies(ext);
    generateRemoteInitEPV(ext);
    generateRemoteConnect(ext);
    generateIHConnect(ext);
    if(!ext.isAbstract()) {  //No constructor for abstract classes
      generateRemoteConstructor(cls);
    }
    generateRMICast(cls);
    generateConnectInternal(cls);
  }

  /**
   * Generate the RStub source for a sidl interface. The source file begins with
   * a banner and include files followed by a declaration of static methods and
   * (for a class) external methods expected in the skeleton file. For classes,
   * the source file then defines a number of functions (cast, delete,
   * initialize EPVs, new, init, and fini). This is the version used for
   * interfaces.
   */
  private void generateSource(Interface ifc, Class anonCls, boolean includes)
      throws CodeGenerationException {
    /*
     * Generate the file banner and include files.
     */
    SymbolID id = anonCls.getSymbolID();

    /*if (!anonCls.isAbstract()) {
      lookupSymbolIDs();
      d_writer.generateInclude(C.getHeaderFile(d_baseClass), true);
      d_writer.generateInclude(C.getHeaderFile(d_classInfo), true);
      }*/
    if(includes) {
      generateIncludes(anonCls);
      generateStaticVariables(anonCls);
    }
    /*
     * Generate internal static variables and anonClsernal references to be
     * supplied by the skeleton file.
     */
    

    /*
     * Generate remote method support for interfaces and classes.
     */
    IOR.generateCastFunction(anonCls, "self", d_writer, true,true);
    //generateRemoteCastFunction(anonCls, ifc);
    generateRemoteDeleteFunction(anonCls);
    generateRemoteGetURLFunction(anonCls);
    generateRemoteAddRefFunction(anonCls);
    generateIsLocalIsRemote(anonCls);
    generateRemoteExecFunction(anonCls);
    generateRemoteMethodBodies(anonCls);
    generateRemoteInitEPV(anonCls);
    generateRemoteConnect(ifc, anonCls);
    generateIHConnect(ifc, anonCls);
    generateRMICast(ifc);
    generateConnectInternal(ifc);
  }

  /**
   * Generate a single line comment. This is called out as a separate method to
   * make the code formatting below a little prettier.
   */
  private void comment(String s) {
    d_writer.writeCommentLine(s);
  }

  private void generateIncludes(Extendable ext) throws CodeGenerationException {
    SymbolID id = ext.getSymbolID();
    SymbolTable table = SymbolTable.getInstance();
    
    d_writer.printlnUnformatted("#include <stdlib.h>");
    d_writer.printlnUnformatted("#include <string.h>");
    // if (!ext.isAbstract()) {
    lookupSymbolIDs();
    d_writer.generateInclude(C.getHeaderFile(d_baseClass), true);
    d_writer.generateInclude(C.getHeaderFile(d_classInfo), true);
      //}

    d_writer.generateInclude(C.getHeaderFile(table.lookupSymbol(
        "sidl.rmi.ProtocolFactory").getSymbolID()), true);
    d_writer.generateInclude(C.getHeaderFile(table.lookupSymbol(
        "sidl.rmi.InstanceRegistry").getSymbolID()), true);
    d_writer.generateInclude(C.getHeaderFile(table.lookupSymbol(
        "sidl.rmi.InstanceHandle").getSymbolID()), true);
    d_writer.generateInclude(C.getHeaderFile(table.lookupSymbol(
        "sidl.rmi.Invocation").getSymbolID()), true);
    d_writer.generateInclude(C.getHeaderFile(table.lookupSymbol(
        "sidl.rmi.Response").getSymbolID()), true);
    d_writer.generateInclude(C.getHeaderFile(table.lookupSymbol(
        "sidl.rmi.ServerRegistry").getSymbolID()), true);
    d_writer.generateInclude(C.getHeaderFile(table.lookupSymbol(
        "sidl.rmi.ConnectRegistry").getSymbolID()), true);
    d_writer.generateInclude(C.getHeaderFile(table.lookupSymbol(
        "sidl.io.Serializable").getSymbolID()), true);

    d_writer.printlnUnformatted("#include \"sidl_Exception.h\"");

    d_writer.println();

    d_writer.printlnUnformatted("#ifndef " + C.NULL);
    d_writer.printlnUnformatted("#define " + C.NULL + " 0");
    d_writer.printlnUnformatted("#endif");
    d_writer.println();

    String my_mutex = IOR.getSymbolName(id) + "__mutex";
    String lock_name = IOR.getLockStaticGlobalsMacroName();
    String unlock_name = IOR.getUnlockStaticGlobalsMacroName();
    String have_lock = IOR.getHaveLockStaticGlobalsMacroName();
    d_writer.printlnUnformatted("#include \"sidl_thread.h\"");
    d_writer.printlnUnformatted("#ifdef HAVE_PTHREAD");
    d_writer.printlnUnformatted("static struct sidl_recursive_mutex_t "
        + my_mutex + "= SIDL_RECURSIVE_MUTEX_INITIALIZER;");
    d_writer.printlnUnformatted("#define " + lock_name
        + " sidl_recursive_mutex_lock( &" + my_mutex + " )");
    d_writer.printlnUnformatted("#define " + unlock_name
        + " sidl_recursive_mutex_unlock( &" + my_mutex + " )");
    d_writer.printlnUnformatted("/* #define " + have_lock
        + " (sidl_recursive_mutex_trylock( &" + my_mutex + " )==EDEADLOCK) */");
    d_writer.printlnUnformatted("#else");
    d_writer.printlnUnformatted("#define " + lock_name);
    d_writer.printlnUnformatted("#define " + unlock_name);
    d_writer.printlnUnformatted("/* #define " + have_lock + " (1) */");
    d_writer.printlnUnformatted("#endif");
    d_writer.println();

  }

  /**
   * Generate the static variables used to store the EPVs and also the
   * initialization flags for the EPVs. For interfaces, we only need to generate
   * a remote EPV structure. Classes require EPVs for their static methods (if
   * present), standard methods, remote methods, and old, new, and remote
   * versions for all parent classes and interfaces.
   */
  private void generateStaticVariables(Extendable ext) {
    comment("Static variables to hold version of IOR");
    d_writer.println("static const int32_t s_IOR_MAJOR_VERSION = "
        + IOR.MAJOR_VERSION + ";");
    d_writer.println("static const int32_t s_IOR_MINOR_VERSION = "
        + IOR.MINOR_VERSION + ";");
    d_writer.println();

    if (!ext.isAbstract()) {
      /*
       * TODO: ClassInfo and RMI? comment("Static variable to hold shared
       * ClassInfo interface."); d_writer.println("static " +
       * C.getObjectName(d_classInfo) + " s_classInfo = " + C.NULL + ";");
       * d_writer.println("static int s_classInfo_init = 1;");
       * d_writer.println();
       */
    }

    comment("Static variables for managing EPV initialization.");

    /*
     * Output the initialization flags for the EPV structures.
     */
    d_writer.println("static int s_remote_initialized = 0;");
    d_writer.println();

    /*
     * Output the EPV, remote EPV, and static EPV for this object.
     */
    IOR.generateStaticEPVVariables(d_writer, ext, false, true, IOR.SET_PUBLIC);

    /*
     * If the object is a class, then collect all the parents in a set and
     * output EPV structures for the parents.
     */
    if (!ext.isInterface()) {
      Class cls = (Class) ext;
      Set parents = Utilities.getAllParents(cls);
      Set new_interfaces = Utilities.getUniqueInterfaceIDs(cls);

      if (!parents.isEmpty()) {
        List sorted = Utilities.sort(parents);
        for (Iterator i = sorted.iterator(); i.hasNext();) {
          SymbolID p_id = (SymbolID) i.next();
          String p_epv = "static " + IOR.getEPVName(p_id);
          boolean is_old = !new_interfaces.contains(p_id);
          String p_epvStr = (is_old ? p_epv + "  " : p_epv + " ");

          d_writer.print(p_epvStr);
          d_writer.println(IOR.getStaticEPVVariable(p_id, IOR.EPV_REMOTE,
              IOR.SET_PUBLIC)
              + ";");
          d_writer.println();
        }
      }
    }
  }

  /**
   * Recursively generate the cast logic for a class and its interfaces. This
   * routine first generates the check against the class type and then against
   * all newly defined interfaces in the class. Finally, it recurses on the
   * class parent (if it exists).
   */
  private void generateCastLogic(Class cls, int level) {
    String self = "s" + String.valueOf(level);

    /*
     * Check the cast against the class type.
     */
    if (level > 0) {
      d_writer.print("} else ");
    }
    d_writer.println("if (!strcmp(name, \"" + cls.getFullName() + "\")) {");
    d_writer.tab();
    d_writer.println("cast = (void*) " + self + ";");
    d_writer.backTab();

    /*
     * Check the cast against all newly defined interfaces in this class.
     */
    List interfaces = Utilities.sort(Utilities.getUniqueInterfaceIDs(cls));
    for (Iterator i = interfaces.iterator(); i.hasNext();) {
      SymbolID id = (SymbolID) i.next();
      String fn = id.getFullName();
      String lower = IOR.getSymbolName(id).toLowerCase();
      d_writer.println("} else if (!strcmp(name, \"" + fn + "\")) {");
      d_writer.tab();
      d_writer.println("cast = (void*) &" + self + "->d_" + lower + ";");
      d_writer.backTab();
    }

    /*
     * If there is a parent class, then recursively call the cast function on
     * the parent class. Otherwise, close the if test block.
     */
    Class parent = cls.getParentClass();
    if (parent != null) {
      generateCastLogic(parent, level + 1);
    } else {
      d_writer.println("}");
    }
  }

  /**
   * Returns the name of the built-in method, prepending "ior_" and the name of
   * the symbol.
   */
  private String getRemoteMethodName(SymbolID id, String name) {
    return "remote_" + IOR.getSymbolName(id) + '_' + name;
  }

  /**
   * Generate a return string for the specified SIDL type. Most of the SIDL
   * return strings are listed in the static structures defined at the start of
   * this class. Symbol types and array types require special processing.
   */
  private static String getReturnString(Type type)
      throws CodeGenerationException {
    return IOR.getReturnString(type, true, false);
  }

  /**
   * Generate the remote cast method body.
   */
  private void generateRemoteCastFunction(Class cls, Interface ifc) {
    comment("REMOTE CAST: dynamic type casting for remote objects.");

    SymbolID id = cls.getSymbolID();
    String name = IOR.getSymbolName(id);

    d_writer.println("static void* remote_" + name + '_' + s_castBuiltin + '(');
    d_writer.println(IOR.getObjectName(id) + "* self,");
    d_writer.println("const char* name,");
    d_writer.println("sidl_BaseInterface* _ex)");
    d_writer.println("{");
    d_writer.tab();

    /*
     * Output parent pointers to simplify access to parent classes. Recursively
     * generate the casting checks for this class and all parent classes.
     */
    d_writer.println("void* cast = " + C.NULL + ";");
    d_writer.println();

    declareParentSelf(cls, 0);
    generateParentSelf(cls, 0, 0);
    d_writer.println();

    generateCastLogic(cls, 0);
    // A little hack to get the correct isType name if we're generating
    // inside an anonymous class. There should be a smarter way....
    //    if (ifc != null) {
    d_writer.println("else if ((*self->d_epv->f_isType)(self,name, _ex)) {");

    d_writer.tab();
    d_writer.println("void* (*func)(struct sidl_rmi_InstanceHandle__object*) = ");
    d_writer.tab();
    d_writer.println("(void* (*)(struct sidl_rmi_InstanceHandle__object*)) ");
    d_writer.println("sidl_rmi_ConnectRegistry_getConnect(name, _ex);");
    d_writer.backTab();

    d_writer.println("cast =  (*func)(((" + IOR.getRemoteStructName(id)
        + "*)self->d_data)->d_ih);");
    d_writer.backTab();
    d_writer.println("}");

    d_writer.println();

    /*
     * Return the cast pointer and close the function body.
     */
    d_writer.println("return cast;");
    d_writer.backTab();

    d_writer.backTab();
    d_writer.println("}");
    d_writer.println();
  }

  /**
   * Generate the remote delete method body
   */
  private void generateRemoteDeleteFunction(Extendable ext) {
    comment("REMOTE DELETE: call the remote destructor for the object.");

    SymbolID id = ext.getSymbolID();
    String name = IOR.getSymbolName(id);

    d_writer
        .println("static void remote_" + name + '_' + s_deleteBuiltin + '(');
    d_writer.tab();
    if (ext.isInterface()) {
      d_writer.println("void* self,");
    } else {
      d_writer.println(IOR.getObjectName(id) + "* self,");
    }
    d_writer.println("sidl_BaseInterface* _ex)");
    d_writer.backTab();
    d_writer.println("{");
    d_writer.tab();
    d_writer.println("free((void*) self);");
    d_writer.backTab();
    d_writer.println("}");
    d_writer.println();
  }

  /**
   * TODO: Generate the remote exec method body
   */
  private void generateRemoteExecFunction(Extendable ext)
      throws CodeGenerationException {
    comment("REMOTE EXEC: call the exec function for the object.");

    SymbolID id = ext.getSymbolID();
    String name = IOR.getSymbolName(id);
    String object = IOR.getObjectName(id);
    Method m_exec = IOR.getBuiltinMethod(IOR.EXEC, id, false);

    d_writer.println("static void remote_" + name + '_' + s_execBuiltin + '(');
    d_writer.tab();
    List args = m_exec.getArgumentList();

    d_writer.print(object + "* self,");

    for (Iterator a = args.iterator(); a.hasNext();) {
      Argument arg = (Argument) a.next();
      d_writer.print(IOR.getArgumentWithFormal(arg, false, false, true));
      d_writer.println(",");
    }
    d_writer.println("sidl_BaseInterface* _ex)");
    d_writer.backTab();
    d_writer.println("{");
    d_writer.tab();
    d_writer.backTab();
    d_writer.println("}");
    d_writer.println();
  }

  private void generateRemoteGetURLFunction(Extendable ext)
      throws CodeGenerationException {
    comment("REMOTE GETURL: call the getURL function for the object.");

    SymbolID id = ext.getSymbolID();
    String name = IOR.getSymbolName(id);
    String object = IOR.getObjectName(id);
    Method m_getURL = IOR.getBuiltinMethod(IOR.GETURL, id, false);

    d_writer.println("static char* remote_" + name + '_' + s_getURLBuiltin
        + '(');
    d_writer.tab();
    List args = m_getURL.getArgumentList();

    d_writer.print(object + "* self, ");
    for (Iterator a = args.iterator(); a.hasNext();) {
      Argument arg = (Argument) a.next();
      d_writer.print(IOR.getArgumentWithFormal(arg, false, false, true));
      d_writer.println(",");
    }
    d_writer.println("sidl_BaseInterface* _ex)");
    d_writer.backTab();
    d_writer.println("{");
    d_writer.tab();
    d_writer.println("sidl_rmi_InstanceHandle conn = " + "(("
        + IOR.getRemoteStructName(id) + "*)self->d_data)->d_ih;");

    d_writer.println("if(conn != NULL) {");
    d_writer.tab();
    d_writer.println("return sidl_rmi_InstanceHandle_getObjectURL(conn, _ex);");
    d_writer.backTab();
    d_writer.println("}");
    d_writer.println("return NULL;");
    d_writer.backTab();
    d_writer.println("}");
    d_writer.println();
  }

  // TODO: Don't forget to add exception handling!
  private void generateRemoteAddRefFunction(Extendable ext)
      throws CodeGenerationException {
    comment("REMOTE ADDREF: For internal babel use only! Remote addRef.");

    SymbolID id = ext.getSymbolID();
    String name = IOR.getSymbolName(id);
    String object = IOR.getObjectName(id);
    Method m_raddRef = IOR.getBuiltinMethod(IOR.RADDREF, id, false);
    d_writer.println("static void remote_" + name + '_' + s_raddRefBuiltin
        + '(');
    d_writer.tab();
    List args = m_raddRef.getArgumentList();

    d_writer.print(object + "* self,");

    for (Iterator a = args.iterator(); a.hasNext();) {
      Argument arg = (Argument) a.next();
      d_writer.print(IOR.getArgumentWithFormal(arg, false, false, true));
      d_writer.println(",");
    }
    d_writer.println("sidl_BaseInterface* _ex)");
    d_writer.backTab();
    d_writer.println("{");
    d_writer.tab();
    d_writer.println("sidl_BaseException netex = NULL;");
    comment("initialize a new invocation");

    d_writer.println("sidl_rmi_InstanceHandle _conn = " + "(("
        + IOR.getRemoteStructName(id) + "*)self->d_data)->d_ih;");
    d_writer.println("sidl_rmi_Response _rsvp = " + C.NULL + ";");
    d_writer.println("sidl_rmi_Invocation _inv = "
        + "sidl_rmi_InstanceHandle_createInvocation( _conn, "
        + "\"addRef\", _ex ); SIDL_CHECK(*_ex);");

    comment("send actual RMI request");

    d_writer.println("_rsvp = sidl_rmi_Invocation_invokeMethod(_inv,"
        + "_ex);SIDL_CHECK(*_ex);");

    comment("Check for exceptions");
    d_writer.println("netex = sidl_rmi_Response_getExceptionThrown(_rsvp, _ex);");
    d_writer.println("if(netex != NULL) {");
    d_writer.tab();
    d_writer.println("sidl_BaseInterface throwaway_exception = NULL;");
    d_writer.println("*_ex = (sidl_BaseInterface) sidl_BaseInterface__rmicast(netex,&throwaway_exception);");
    d_writer.println("return;");
    d_writer.backTab();
    d_writer.println("}");
    d_writer.println();

    comment("cleanup and return");
    d_writer.println("sidl_rmi_Invocation_deleteRef(_inv,_ex);");
    d_writer.println("sidl_rmi_Response_deleteRef(_rsvp,_ex);");
    d_writer.println("EXIT:");
    d_writer.println("return;");
    d_writer.backTab();
    d_writer.println("}");
    d_writer.println();
  }

  /**
   * @param cls
   */
  private void generateIsLocalIsRemote(Extendable ext) {
    SymbolID id = ext.getSymbolID();
    String my_symbolName = IOR.getSymbolName(id);

    comment("REMOTE ISREMOTE: returns true if this object is Remote (it is).");
    d_writer.println("static sidl_bool");
    d_writer.println("remote_" + IOR.getRemoteIsRemoteName(id) + "(");
    d_writer.println("    struct " + my_symbolName + "__object* self, ");
    d_writer.println("    sidl_BaseInterface *_ex) {");
    d_writer.tab();

    d_writer.println("return TRUE;");
    d_writer.backTab();
    d_writer.println("}");
    d_writer.println("");
  }

  /**
   * Generate remote method bodies for all non-static user-defined methods in an
   * extendable.
   * 
   * @param ext
   *          The extendable to generate methods for
   * @throws CodeGenerationException
   * @see #generateSingleRemoteMethodBody(Extendable, Method)
   */
  private void generateRemoteMethodBodies(Extendable ext)
      throws CodeGenerationException {
    Iterator m = ext.getNonstaticMethods(true).iterator();
    while (m.hasNext()) {
      Method method = (Method) m.next();
      generateSingleRemoteMethodBody(ext, method);
      if (method.getCommunicationModifier() == Method.NONBLOCKING) {
        generateSingleRemoteMethodBody(ext, method.spawnNonblockingSend());
        generateSingleRemoteMethodBody(ext, method.spawnNonblockingRecv());
      }
    }
  }

  /**
   * Generate the body for a single remote user-defined method belonging to an
   * extendable
   * 
   * @param ext
   *          The extendable owning the method
   * @param method
   *          The method in particular
   * @throws CodeGenerationException
   */
  private void generateSingleRemoteMethodBody(Extendable ext, Method method)
      throws CodeGenerationException {
    SymbolID id = ext.getSymbolID();
    String name = IOR.getSymbolName(id);
    String method_name = method.getLongMethodName();
    Type returnType = method.getReturnType();
    boolean hasThrows = !method.getThrows().isEmpty();

    comment("REMOTE METHOD STUB:" + method_name);

    /* Write the method signature. */
    d_writer.println("static " + getReturnString(returnType));
    d_writer.println("remote_" + name + "_" + method_name + "(");
    d_writer.tab();
    C.generateArgumentList(d_writer, "self ", ext.isInterface(), id, method,
        true, true, true, hasThrows, false, false, false);
    d_writer.println(")");
    d_writer.backTab();

    /* Write the method body. */
    d_writer.println("{");
    d_writer.tab();

    d_writer.println(RMI.LangSpecificInit()+";");

    if (method_name.compareTo("addRef") == 0) {
      remoteAddRefFragment(id);
    } else if (method_name.compareTo("deleteRef") == 0) {
      remoteDelRefFragment(id);
    } else {
      remoteMethodGenericFragment(ext, method);
    }

    /* close the body out */
    d_writer.backTab();
    d_writer.println("}");
    d_writer.println();
  }

  /**
   * Generate the fragment of code specific to a remote addRef method
   * 
   * @param id
   * @see #generateSingleRemoteMethodBody(Extendable, Method)
   */
  private void remoteAddRefFragment(SymbolID id) {
    d_writer.println(IOR.getRemoteStructName(id) + "* r_obj = ("
        + IOR.getRemoteStructName(id) + "*)self->d_data;");
    d_writer.println(IOR.getLockStaticGlobalsMacroName() + ";");
    d_writer.println("r_obj->d_refcount++;");
    d_writer.println(IOR.getUnlockStaticGlobalsMacroName() + ";");
  }

  /**
   * Generate the fragment of code specific to a remote deleteRef method
   * 
   * @param id
   * @see #generateSingleRemoteMethodBody(Extendable, Method)
   */
  private void remoteDelRefFragment(SymbolID id) {
    d_writer.println(IOR.getRemoteStructName(id) + "* r_obj = ("
        + IOR.getRemoteStructName(id) + "*)self->d_data;");
    d_writer.println(IOR.getLockStaticGlobalsMacroName() + ";");
    d_writer.println("r_obj->d_refcount--;");
    d_writer.println("if(r_obj->d_refcount == 0) {");
    d_writer.tab();
    d_writer.println("sidl_rmi_InstanceHandle_deleteRef(r_obj->d_ih, _ex);");
    d_writer.println("free(r_obj);");
    d_writer.println("free(self);");
    d_writer.backTab();
    d_writer.println("}");
    d_writer.println(IOR.getUnlockStaticGlobalsMacroName() + ";");
  }

  /**
   * Generates the packing and unpacking commands for the remote case of
   * general, user-defined methods in an Extendable
   * 
   * @param ext
   * @param method
   * @throws CodeGenerationException
   */
  private void remoteMethodGenericFragment(Extendable ext, Method method)
      throws CodeGenerationException {
    SymbolID id = ext.getSymbolID();
    String blocking_method_name = method.getBlockingMethodName();
    List args = method.getArgumentList();
    Type returnType = method.getReturnType();
    int commType = method.getCommunicationModifier();
    boolean outDisallowed = (commType == Method.NONBLOCKING_SEND || commType == Method.ONEWAY);
    boolean inDisallowed = (commType == Method.NONBLOCKING_RECV);

    comment("initialize a new invocation");
    d_writer.println("sidl_rmi_InstanceHandle _conn = " + "(("
        + IOR.getRemoteStructName(id) + "*)self->d_data)->d_ih;");

    if (!outDisallowed) {

      d_writer.println("sidl_BaseException _be = " + C.NULL + ";");
      d_writer.println("sidl_rmi_Response _rsvp = " + C.NULL + ";");

      // Generate the url holders for any out or inout object arguments
      for (Iterator a = args.iterator(); a.hasNext();) {
        Argument arg = (Argument) a.next();
        if ((arg.getType().getDetailedType() == Type.CLASS || arg.getType()
            .getDetailedType() == Type.INTERFACE)
            && arg.getMode() != Argument.IN) {
          if(arg.isCopy()) {
            d_writer.println("sidl_io_Serializable " +arg.getFormalName()+ 
                             RMI.getSerializableExt()+" = "+C.NULL+ ";");
          } else {
            // We need a string to unserialize the url into
            d_writer.println(IOR.getReturnString(new Type(Type.STRING)) + " "
              + arg.getFormalName() + "_str= NULL;");
          }
        }
        if (arg.getType().getDetailedType() == Type.ENUM 
            && arg.getMode() != Argument.IN) {
          d_writer.println(IOR.getReturnString(new Type(Type.INT)) + " "
                           + arg.getFormalName() + "_tmp= 0;");
        }
      }
    }

    if (outDisallowed) {
      d_writer.println("struct sidl_rmi_Ticket__object *_retval = NULL;");
    }
    else {
      if ((returnType.getType() != Type.VOID)) {
        if (returnType.getDetailedType() == Type.CLASS
            || returnType.getDetailedType() == Type.INTERFACE) {
          if(method.isReturnCopy()) {
            d_writer.println("sidl_io_Serializable "+RMI.getReturnArgName()+ 
                             RMI.getSerializableExt()+" = "+C.NULL+ ";");
          } else {
            d_writer.println(IOR.getReturnString(new Type(Type.STRING))
                             + RMI.getReturnArgName()+RMI.getStringExt() + " = "+C.NULL+";");
          }
        }
        else if (returnType.getDetailedType() == Type.ENUM) {
          d_writer.println(IOR.getReturnString(new Type(Type.INT)) + " "
                           + RMI.getReturnArgName() + "_tmp = 0;");
        }
        d_writer.println(getReturnString(returnType) + " " +
                         RMI.getReturnArgName() +
                         Utilities.getTypeInitialization(returnType) + ";");
      }
    }
    d_writer.println("sidl_rmi_Invocation _inv = "
        + "sidl_rmi_InstanceHandle_createInvocation( _conn, \""
        + blocking_method_name + "\", _ex ); SIDL_CHECK(*_ex);");
    d_writer.println();

    if (!inDisallowed) {
      comment("pack in and inout arguments");
      for (Iterator a = args.iterator(); a.hasNext();) {
        Argument arg = (Argument) a.next();
        if (arg.getMode() == Argument.IN || arg.getMode() == Argument.INOUT) {
          RMI.packArg(d_writer, "sidl_rmi_Invocation", "_inv", arg, false);
        }
      }
    }

    /* TODO: There may be something smarter we can do for transfering references */
    for (Iterator a = args.iterator(); a.hasNext();) {
      Argument arg = (Argument) a.next();
      if ((arg.getType().getDetailedType() == Type.CLASS || 
           arg.getType().getDetailedType() == Type.INTERFACE)
          && arg.getMode() == Argument.INOUT && !arg.isCopy()) {
        comment("Transfer this reference");
        d_writer.println("if(*" + arg.getFormalName()
            + " && sidl_BaseInterface__isRemote(" + "(sidl_BaseInterface)*"
            + arg.getFormalName() + ", _ex)) {");
        d_writer.tab();
        d_writer.println("SIDL_CHECK(*_ex);");
        d_writer.println("(*((sidl_BaseInterface)*" + arg.getFormalName()
            + ")->d_epv->f__raddRef)(((sidl_BaseInterface)*"
            + arg.getFormalName() + ")->d_object, _ex);SIDL_CHECK(*_ex);");
        d_writer.println("sidl_BaseInterface_deleteRef((sidl_BaseInterface)*"
            + arg.getFormalName() + ", _ex);SIDL_CHECK(*_ex); ");
        d_writer.backTab();
        d_writer.println("}");
      }
    }
    d_writer.println();

    comment("send actual RMI request");
    switch (commType) {
    case Method.ONEWAY:
      d_writer
          .println("sidl_rmi_Invocation_invokeOnewayMethod(_inv, _ex);SIDL_CHECK(*_ex);");
      break;
    case Method.NONBLOCKING_SEND:
      d_writer
          .println("_retval = sidl_rmi_Invocation_invokeNonblocking(_inv, _ex);SIDL_CHECK(*_ex);");
      break;
    case Method.NONBLOCKING_RECV:
      d_writer.println("_rsvp = sidl_rmi_Ticket_getResponse(_ticket, _ex);SIDL_CHECK(*_ex);");
      break;

   case Method.LOCAL:
      //TODO: Throw an exception here or something.  (Fall through to
      //default for now)  

    case Method.NONBLOCKING:
    /*
     * Intentional fallthrough. Nonblocking method called in blocking fashion is
     * same as NORMAL
     */
    case Method.NORMAL:
    default:
      d_writer
          .println("_rsvp = sidl_rmi_Invocation_invokeMethod(_inv, _ex);SIDL_CHECK(*_ex);");
      break;
    }
    d_writer.println();
    if (!outDisallowed ) { 
        d_writer.println("_be = sidl_rmi_Response_getExceptionThrown(_rsvp, _ex);");
        d_writer.println("if(_be != NULL) {");
        d_writer.tab();
        d_writer.println("sidl_BaseInterface throwaway_exception = NULL;");
        d_writer.println("*_ex = (sidl_BaseInterface) sidl_BaseInterface__rmicast(_be,&throwaway_exception);");
        d_writer.println("goto EXIT;");
        d_writer.backTab();
        d_writer.println("}");
        d_writer.println();
    }

    if ((returnType.getType() != Type.VOID)
        && (method.getCommunicationModifier() != Method.NONBLOCKING_SEND)) {
      // nothing to process for void returns
      // and return type for nonblocking send is a special case
      // ( sidl.rmi.Ticket is generated as a local instance by the protocol)
      comment("extract return value");
      RMI
          .unpackType(d_writer, ext, "sidl_rmi_Response", "_rsvp", returnType,
              "_retval", "_retval", Argument.IN, method.isReturnCopy(), true,
              false);
      d_writer.println();
    }

    if (!outDisallowed) {
      comment("unpack out and inout arguments");
      for (Iterator a = args.iterator(); a.hasNext();) {
        Argument arg = (Argument) a.next();
        if (arg.getMode() != Argument.IN) {
          RMI
              .unpackArg(d_writer, ext, "sidl_rmi_Response", "_rsvp", arg,
                  false);
        }
      }
      d_writer.println();
    }

    comment("cleanup and return");
    d_writer.println("sidl_rmi_Invocation_deleteRef(_inv, _ex); SIDL_CHECK(*_ex);");
    if (!outDisallowed) {
      d_writer.println("sidl_rmi_Response_deleteRef(_rsvp, _ex); SIDL_CHECK(*_ex);");
    }
    d_writer.println("EXIT:");
    if (returnType.getType() != Type.VOID) {
      d_writer.println("return _retval;");
    } else {
      d_writer.println("return;");
    }
  }

  /**
   * Generate the function that initializes the method entry point vector for a
   * remote instance. This method performs two basic functions. First, it
   * initializes the remote EPV with pointers to the remote stubs in the IOR
   * file. Second, it sets interface and parent EPV function pointers using the
   * pointer values generated in the first step.
   */
  private void generateRemoteInitEPV(Extendable ext)
      throws CodeGenerationException {
    comment("REMOTE EPV: create remote entry point vectors (EPVs).");

    /*
     * Decleare the method signature and open the implementation body.
     */
    SymbolID id = ext.getSymbolID();
    String name = IOR.getSymbolName(id);

    d_writer.println("static void " + name + "__init_remote_epv(void)");
    d_writer.println("{");
    d_writer.tab();
    comment("assert( " + IOR.getHaveLockStaticGlobalsMacroName() + " );");

    /*
     * Output entry point vectors aliases for each parent class and interface as
     * well as a special one for the current object.
     */
    List parents = null;
    if (ext.isInterface()) {
      d_writer.print(IOR.getEPVName(id) + "*");
      d_writer.println(" epv = &"
          + IOR.getStaticEPVVariable(id, IOR.EPV_REMOTE, IOR.SET_PUBLIC) + ";");
      d_writer.println();
    } else {
      parents = Utilities.sort(Utilities.getAllParents((Class) ext));
      aliasEPVs((Class) ext, parents, true);
    }

    /*
     * Generate a list of the nonstatic methods in the class and get the width
     * of every method name.
     */
    List methods = (List) ext.getNonstaticMethods(true);
    int mwidth = Math.max(Utilities.getWidth(methods), s_longestBuiltin)
        + IOR.getVectorEntry("").length() + IOR.GENERIC_POST_SUFFIX.length();

    /*
     * Output standard methods.
     */
    boolean doHooks = false;

    for (int j = 0; j < IOR.CLASS_BUILTIN_METHODS; j++) {
      if (IOR.isBuiltinBasic(j) || ((j == IOR.HOOKS) && doHooks)) {
        String mname = IOR.getBuiltinName(j);
        d_writer.print("epv->");
        d_writer.printAligned(IOR.getVectorEntry(mname), mwidth);
        if ((j == IOR.CONSTRUCTOR) || (j == IOR.DESTRUCTOR)
            || (j == IOR.CHECK_ERROR)) {
          d_writer.println(" = " + C.NULL + ";");
        } else {
          d_writer.println(" = " + getRemoteMethodName(id, mname) + ';');
        }
      }
    }

    /*
     * Iterate through all of the nonstatic methods. Assign them to the remote
     * stub function.
     */
    for (Iterator i = methods.iterator(); i.hasNext();) {
      Method method = (Method) i.next();
      assignRemoteEPVPointer(name, method, mwidth, doHooks);
      if (method.getCommunicationModifier() == Method.NONBLOCKING) {
        assignRemoteEPVPointer(name, method.spawnNonblockingSend(), mwidth,
            doHooks);
        assignRemoteEPVPointer(name, method.spawnNonblockingRecv(), mwidth,
            doHooks);
      }
    }

    /*
     * If this is a class, then iterate through all parent EPVs and set each of
     * the function pointers to use the corresponding function in the class EPV
     * structure.
     */
    if (parents != null) {
      d_writer.println();
      copyEPVs(parents);
    }
    d_writer.println("s_remote_initialized = 1;");

    /*
     * Close the function definition.
     */
    d_writer.backTab();
    d_writer.println("}");
    d_writer.println();
  }

  /**
   * Generate a single assignment in initialization of the remote EPV pointer
   * 
   * @param name
   *          String name of the extendable
   * @param method
   *          The method in the EPV to initialzie
   * @param mwidth
   *          formatting parameter related to width of longest method name
   * @param doHooks
   *          trigger for extra semantic code generation
   */
  private void assignRemoteEPVPointer(String name, Method method, int mwidth,
      boolean doHooks) {
    String mname = method.getLongMethodName();
    String ename = IOR.getVectorEntry(mname);
    if (doHooks) {
      d_writer.print("epv->");
      d_writer.printAligned(ename + IOR.GENERIC_PRE_SUFFIX, mwidth);
      d_writer.println(" = " + C.NULL + ";");
      d_writer.print("epv->");
      d_writer.printAligned(ename, mwidth);
      d_writer.println(" = remote_" + name + "_" + mname + ";");
      d_writer.print("epv->");
      d_writer.printAligned(ename + IOR.GENERIC_POST_SUFFIX, mwidth);
      d_writer.println(" = " + C.NULL + ";");
    } else {
      d_writer.print("epv->");
      d_writer.printAligned(ename, mwidth);
      d_writer.println(" = remote_" + name + "_" + mname + ";");
    }
  }

  /**
   * Generate remote connect support for classes
   */
  private void generateRemoteConnect(Extendable ext) {
    comment("Create an instance that connects to an "
        + "existing remote object.");

    SymbolID id = ext.getSymbolID();
    String name = IOR.getSymbolName(id);
    String object = C.getSymbolObjectPtr(id);
    String bi_ior = BabelConfiguration.getBaseInterface().replace('.', '_');

    /*
     * Declare the function prototype.
     */
    d_writer.println("static " + object);
    d_writer.println(IOR.getSymbolName(id) + "__remoteConnect"
        + "(const char *url, sidl_bool ar, sidl_BaseInterface *_ex)");
    d_writer.println("{");
    d_writer.tab();

    /*
     * Data
     */
    declareRemoteEPV(ext);
    d_writer.println("sidl_rmi_InstanceHandle instance = NULL;");
    d_writer.println("char* objectID = NULL;");
    d_writer.println("objectID = NULL;");
    d_writer.println("if(url == NULL) {return NULL;}");
    d_writer.println("objectID = sidl_rmi_ServerRegistry_isLocalObject(url, _ex);");
    d_writer.println("if(objectID) {");
    d_writer.tab();
    d_writer.println(bi_ior + " bi = (" + bi_ior
        + ")sidl_rmi_InstanceRegistry_getInstanceString(objectID, _ex); SIDL_CHECK(*_ex);");
    // TODO: REMOVE ADDREF WHEN CAST STARTS ADDREFING
    d_writer.println("if(ar) {");
    d_writer.tab();
    d_writer.println("sidl_BaseInterface_addRef(bi,_ex);SIDL_CHECK(*_ex);");
    d_writer.backTab();
    d_writer.println("}");
    //   comment("return " + name + "__cast(bi);");
    d_writer.println("return " + name + "__rmicast(bi,_ex);SIDL_CHECK(*_ex);");
    d_writer.backTab();
    d_writer.println("}");
    d_writer
        .println("instance = sidl_rmi_ProtocolFactory_connectInstance(url, ar, _ex ); SIDL_CHECK(*_ex);");

    d_writer.println("if ( instance == NULL) { return NULL; }");

    generateRemoteEPV(ext);

    /*
     * Ensure that the EPV has been initialized. d_writer.println(
     * IOR.getLockStaticGlobalsMacroName() + ";" ); d_writer.println("if
     * (!s_remote_initialized) {"); d_writer.tab(); d_writer.println(name +
     * "__init_remote_epv();"); d_writer.backTab(); d_writer.println("}");
     * d_writer.println( IOR.getUnlockStaticGlobalsMacroName() + ";" );
     * d_writer.println();
     */
    d_writer.println();
    d_writer.println("return self;");
    d_writer.println("EXIT:");
    d_writer.println("return NULL;");
    d_writer.backTab();
    d_writer.println("}");
  }

  /**
   * Generate remote connect support for interfaces.
   */
  private void generateRemoteConnect(Interface ifc, Class anonCls) {
    comment("Create an instance that connects to an "
        + "existing remote object.");

    SymbolID ifcID = ifc.getSymbolID();
    String ifcSymName = IOR.getSymbolName(ifcID);
    String ifcObject = C.getSymbolObjectPtr(ifcID);
    String bi_ior = BabelConfiguration.getBaseInterface().replace('.', '_');

    /*
     * Declare the function prototype.
     */
    d_writer.println("static " + ifcObject);
    d_writer.println(ifcSymName + "__remoteConnect"
        + "(const char *url, sidl_bool ar, sidl_BaseInterface *_ex)");
    d_writer.println("{");
    d_writer.tab();

    /*
     * Data
     */
    declareRemoteEPV(anonCls);
    d_writer.println("sidl_rmi_InstanceHandle instance = NULL;");
    d_writer.println("char* objectID = NULL;");
    d_writer.println("objectID = sidl_rmi_ServerRegistry_isLocalObject(url, _ex);");
    d_writer.println("if(objectID) {");
    d_writer.tab();
    d_writer.println(bi_ior + " bi = (" + bi_ior
        + ")sidl_rmi_InstanceRegistry_getInstanceString(objectID, _ex);");
    // TODO: REMOVE ADDREF WHEN CAST STARTS ADDREFING
    d_writer.println("if(ar) {");
    d_writer.tab();
    d_writer.println("sidl_BaseInterface_addRef(bi, _ex);");
    d_writer.backTab();
    d_writer.println("}");
    d_writer.println("return " + ifcSymName + "__rmicast(bi, _ex);");
    d_writer.backTab();
    d_writer.println("}");
    d_writer
        .println("instance = sidl_rmi_ProtocolFactory_connectInstance(url, ar, _ex );");
    d_writer.println("if ( instance == NULL) { return NULL; }");

    generateRemoteEPV(anonCls);

    // Cast it to the interface type before returning.
    d_writer.println();
    d_writer.println("return " + ifcSymName + "__rmicast(self, _ex);");
    d_writer.backTab();
    d_writer.println("}");
  }

  /**
   * Generate connect that takes an already existing instace handle.
   */
  private void generateIHConnect(Extendable ext) {
    comment("Create an instance that uses an already existing ");
    comment("InstanceHandle to connect to an existing remote object.");

    SymbolID id = ext.getSymbolID();

    /*
     * Declare the function prototype.
     */
    d_writer.println("static " + C.getSymbolObjectPtr(id));
    d_writer.println(IOR.getSymbolName(id) + "__IHConnect"
        + "(sidl_rmi_InstanceHandle instance, sidl_BaseInterface *_ex)");
    d_writer.println("{");
    d_writer.tab();

    /*
     * Data
     */
    declareRemoteEPV(ext);
    generateRemoteEPV(ext);

    d_writer.println();
    d_writer.println("sidl_rmi_InstanceHandle_addRef(instance,_ex);SIDL_CHECK(*_ex);");
    d_writer.println("return self;");
    d_writer.println("EXIT:");
    d_writer.println("return NULL;");
    d_writer.backTab();
    d_writer.println("}");
  }

  /**
   * Generate remote connect support for interfaces.
   */
  private void generateIHConnect(Interface ifc, Class anonCls) {
    comment("Create an instance that uses an already existing ");
    comment("InstanceHandel to connect to an existing remote object.");

    SymbolID ifcID = ifc.getSymbolID();
    String ifcSymName = IOR.getSymbolName(ifcID);

    /*
     * Declare the function prototype.
     */
    d_writer.println("static " + C.getSymbolObjectPtr(ifcID));
    d_writer.println(ifcSymName + "__IHConnect"
        + "(sidl_rmi_InstanceHandle instance, sidl_BaseInterface *_ex)");
    d_writer.println("{");
    d_writer.tab();

    /*
     * Data
     */
    declareRemoteEPV(anonCls);
    generateRemoteEPV(anonCls);

    // Cast it to the interface type before returning.
    d_writer.println();
    d_writer.println("sidl_rmi_InstanceHandle_addRef(instance, _ex);");
    d_writer.println("return " + ifcSymName + "__rmicast(self, _ex);");

    d_writer.backTab();
    d_writer.println("}");
  }

  /**
   * Generate remote method support (not yet implemented).
   */
  private void generateRemoteConstructor(Class cls) {
    comment("REMOTE: generate remote instance given URL string.");

    SymbolID id = cls.getSymbolID();
    String idName = id.getFullName();
    String object = IOR.getObjectName(id);

    /*
     * Declare the function prototype.
     */
    d_writer.println("static " + object + "*");
    d_writer.println(IOR.getRemoteName(id)
        + "(const char *url, sidl_BaseInterface *_ex)");
    d_writer.println("{");
    d_writer.tab();

    /*
     * Data
     */
    declareRemoteEPV(cls);
    d_writer.println("sidl_rmi_InstanceHandle instance = "
        + "sidl_rmi_ProtocolFactory_createInstance(url, \"" + idName
        + "\", _ex );");

    d_writer.println("if ( instance == NULL) { return NULL; }");

    generateRemoteEPV(cls);

    d_writer.println();
    d_writer.println("return self;");
    d_writer.backTab();
    d_writer.println("}");
  }

  /*
   * This function is created to seperate the declerations of EPV variables from
   * the initilization of them. Therefore, the names declared here must agree
   * with those used in generateRemoteEPV.
   */
  private void declareRemoteEPV(Extendable ext) {
    SymbolID id = ext.getSymbolID();
    String object = IOR.getObjectName(id);

    d_writer.println(object + "* self;");
    d_writer.println();

    if (!ext.isInterface()) {
      declareParentSelf((Class) ext, 0);
      d_writer.println();
    }
    d_writer.println(IOR.getRemoteStructName(id) + "* r_obj;");

  }

  private void generateRemoteEPV(Extendable ext) {

    SymbolID id = ext.getSymbolID();
    String name = IOR.getSymbolName(id);
    String object = IOR.getObjectName(id);

    /*
     * Allocate the object data.
     */
    d_writer.println("self =");
    d_writer.tab();
    d_writer.println("(" + object + "*) malloc(");
    d_writer.tab();
    d_writer.println("sizeof(" + object + "));");
    d_writer.backTab();
    d_writer.backTab();
    d_writer.println();

    /*
     * Allocate remote object data
     */
    d_writer.println("r_obj =");
    d_writer.tab();
    d_writer.println("(" + IOR.getRemoteStructName(id) + "*) malloc(");
    d_writer.tab();
    d_writer.println("sizeof(" + IOR.getRemoteStructName(id) + "));");
    d_writer.backTab();
    d_writer.backTab();
    d_writer.println();
    d_writer.println("r_obj->d_refcount = 1;");
    d_writer.println("r_obj->d_ih = instance;");

    /*
     * Output parent pointers to simplify access to parent classes.
     */
    if (!ext.isInterface()) {
      generateParentSelf((Class) ext, 0, 0);
      d_writer.println();
    }

    /*
     * Ensure that the EPV has been initialized.
     */
    d_writer.println(IOR.getLockStaticGlobalsMacroName() + ";");
    d_writer.println("if (!s_remote_initialized) {");
    d_writer.tab();
    d_writer.println(name + "__init_remote_epv();");
    d_writer.backTab();
    d_writer.println("}");
    d_writer.println(IOR.getUnlockStaticGlobalsMacroName() + ";");
    d_writer.println();

    /*
     * Recursively modify the parent classes and set up interface pointers.
     */
    if (ext.isInterface()) {
      d_writer.println("self->" + IOR.getEPVVar(IOR.PUBLIC_EPV) + "    = &"
          + IOR.getStaticEPVVariable(id, IOR.EPV_REMOTE, IOR.SET_PUBLIC) + ";");
      // d_writer.println("self->d_object = (void*) instance;");
    } else {
      remoteEPVs((Class) ext, 0);
      d_writer.println("self->d_data = (void*) r_obj;");
    }
  }

  /*
   * This function is created to seperate the declerations of EPV variables from
   * the initilization of them. Therefore, the names declared here must agree
   * with those used in generateParentSelf.
   */
  private void declareParentSelf(Class cls, int level) {
    if (cls != null) {
      SymbolID id = cls.getSymbolID();
      d_writer.print(IOR.getObjectName(id) + "*");
      d_writer.println(" s" + String.valueOf(level) + ";");
      declareParentSelf(cls.getParentClass(), level + 1);
    }
  }

  /**
   * Recursively output self pointers to the sidl objects for this class and its
   * parents. The self pointers are of the form sN, where N is an integer
   * represented by the level argument. If the width is zero, then the width of
   * all parents is generated automatically.
   */
  private void generateParentSelf(Class cls, int level, int width) {
    if (cls != null) {

      /*
       * Calculate the width of this class and all parents for pretty output.
       * Ooh, very pretty.
       */
      if (width == 0) {
        Class parent = cls;
        while (parent != null) {
          int w = IOR.getObjectName(parent.getSymbolID()).length();
          if (w > width) {
            width = w;
          }
          parent = parent.getParentClass();
        }
      }

      /*
       * Now use the width information to print out symbols.
       */
      SymbolID id = cls.getSymbolID();
      // d_writer.printAligned(IOR.getObjectName(id) + "*", width+1);
      d_writer.printAligned("s" + String.valueOf(level) + " = ", width + 1);
      if (level == 0) {
        d_writer.println("self;");
      } else {
        d_writer.print("&s" + String.valueOf(level - 1) + "->d_");
        d_writer.println(IOR.getSymbolName(id).toLowerCase() + ";");
      }
      generateParentSelf(cls.getParentClass(), level + 1, width);
    }
  }

  /*
   * Recursively modify the remote EPVs in parent classes and set up the
   * interfaces.
   */
  private void remoteEPVs(Class cls, int level) {
    if (cls != null) {
      String epvVar = IOR.getEPVVar(IOR.SET_PUBLIC);
      remoteEPVs(cls.getParentClass(), level + 1);

      /*
       * Update the EPVs for all of the new interfaces in this particular class.
       */
      String self = "s" + String.valueOf(level);
      List ifce = Utilities.sort(Utilities.getUniqueInterfaceIDs(cls));
      for (Iterator i = ifce.iterator(); i.hasNext();) {
        SymbolID id = (SymbolID) i.next();
        String n = IOR.getSymbolName(id).toLowerCase();
        d_writer.println(self + "->d_" + n + "." + epvVar + "    = &"
            + IOR.getStaticEPVVariable(id, IOR.EPV_REMOTE, IOR.SET_PUBLIC)
            + ";");
        d_writer.println(self + "->d_" + n + ".d_object = (void*) " + "self;");
        d_writer.println();
      }

      /*
       * Modify the class entry point vector.
       */
      d_writer.println(self + "->d_data = (void*) r_obj;"); // used to be:
      // instance
      d_writer.println(self
          + "->"
          + epvVar
          + "  = &"
          + IOR.getStaticEPVVariable(cls.getSymbolID(), IOR.EPV_REMOTE,
              IOR.SET_PUBLIC) + ";");
      d_writer.println();
    }
  }

  /**
   * Generate the entry point vector alias for each parent class and each
   * interface as well as a special one for the current object.
   */
  private void aliasEPVs(Class cls, Collection parents, boolean remote) {
    /*
     * Get the width of the symbols for pretty printing.
     */
    SymbolID id = cls.getSymbolID();
    String epv = IOR.getEPVName(id);
    int epvType = remote ? IOR.EPV_REMOTE : IOR.EPV_NEW;

    int width = epv.length() + 1;
    int w = Utilities.getWidth(parents) + "struct __epv*".length();
    if (w > width) {
      width = w;
    }

    /*
     * Output the EPV pointer for this class and its class and interface
     * parents.
     */
    d_writer.printAligned(epv + "*", width);
    d_writer.println(" epv = &"
        + IOR.getStaticEPVVariable(id, epvType, IOR.SET_PUBLIC) + ";");
    int e = 0;
    for (Iterator i = parents.iterator(); i.hasNext();) {
      SymbolID sid = (SymbolID) i.next();
      d_writer.printAligned(IOR.getEPVName(sid) + "*", width);
      d_writer.printAligned(" e" + String.valueOf(e++), 4);
      d_writer.println(" = &"
          + IOR.getStaticEPVVariable(sid, epvType, IOR.SET_PUBLIC) + ";");
    }
    d_writer.println();
  }

  /*
   * Copy EPV function pointers from the most derived EPV data structure into
   * all parent class and interface EPVs.
   */
  private void copyEPVs(Collection parents) throws CodeGenerationException {
    int e = 0;
    for (Iterator i = parents.iterator(); i.hasNext();) {
      /*
       * Extract information about the parent extendable object. Generate a list
       * of the nonstatic methods and calculate the width of every method name.
       */
      SymbolID id = (SymbolID) i.next();
      Extendable ext = (Extendable) Utilities.lookupSymbol(id);

      List methods = (List) ext.getNonstaticMethods(true);
      int mwidth = Math.max(Utilities.getWidth(methods), s_longestBuiltin)
          + IOR.getVectorEntry("").length();

      /*
       * Calculate the "self" pointer for the cast function.
       */
      String self = null;
      if (ext.isInterface()) {
        self = "void*";
      } else {
        self = IOR.getObjectName(id) + "*";
      }

      /*
       * Generate the assignments for the cast and delete functions.
       */
      String estring = "e" + String.valueOf(e) + "->";
      String vecEntry = IOR.getVectorEntry(s_castBuiltin);

      d_writer.print(estring);
      d_writer.printAligned(vecEntry, mwidth);
      d_writer.print(" = (void* (*)(" + self);
      d_writer.print(",const char*,sidl_BaseInterface*)) epv->");
      d_writer.print(vecEntry);
      d_writer.println(";");

      vecEntry = IOR.getVectorEntry(s_deleteBuiltin);
      d_writer.print(estring);
      d_writer.printAligned(vecEntry, mwidth);
      d_writer.println(" = (void (*)(" + self + ",sidl_BaseInterface*)) epv->" + vecEntry + ";");

      vecEntry = IOR.getVectorEntry(s_getURLBuiltin);
      d_writer.print(estring);
      d_writer.printAligned(vecEntry, mwidth);
      d_writer.println(" = (char* (*)(" + self + ",sidl_BaseInterface*)) epv->" + vecEntry + ";");

      vecEntry = IOR.getVectorEntry(s_raddRefBuiltin);
      d_writer.print(estring);
      d_writer.printAligned(vecEntry, mwidth);
      d_writer.println(" = (void (*)(" + self + ",sidl_BaseInterface*)) epv->" + vecEntry + ";");

      vecEntry = IOR.getVectorEntry(s_isRemoteBuiltin);
      d_writer.print(estring);
      d_writer.printAligned(vecEntry, mwidth);
      d_writer.println(" = (sidl_bool (*)(" + self + ",sidl_BaseInterface*)) epv->" + vecEntry
          + ";");

      /*
       * Generate the assignment for exec function
       */
      vecEntry = IOR.getVectorEntry(s_execBuiltin);
      Method m_exec = IOR.getBuiltinMethod(IOR.EXEC, id, false);
      String ename = m_exec.getLongMethodName();
      d_writer.print(estring);
      d_writer.printAligned(IOR.getVectorEntry(ename), mwidth);
      d_writer.print(" = " + IOR.getCast(m_exec, self));
      d_writer.println(" epv->" + IOR.getVectorEntry(ename) + ";");

      /*
       * Iterate over all methods in the EPV and set the method pointer.
       */
      for (Iterator j = methods.iterator(); j.hasNext();) {
        Method method = (Method) j.next();
        String mname = method.getLongMethodName();
        d_writer.print(estring);
        d_writer.printAligned(IOR.getVectorEntry(mname), mwidth);
        d_writer.print(" = " + IOR.getCast(method, self));
        d_writer.println(" epv->" + IOR.getVectorEntry(mname) + ";");
      }

      d_writer.println();
      e++;
    }
  }

  public static void generateExternalSignature(LanguageWriter lw, Symbol sym,
      String terminator) {
    final SymbolID id = sym.getSymbolID();
    lw.beginBlockComment(false);
    lw.println("This function returns a pointer to a static structure of");
    lw.println("pointers to function entry points.  Its purpose is to "
        + "provide");
    lw.println("one-stop shopping for loading DLLs.");
    lw.endBlockComment(false);
    lw.println("const " + IOR.getExternalName(id) + "*");
    lw.println(IOR.getExternalFunc(id) + "(void)" + terminator);
  }

  public void generateRMICast(Extendable ext) 
    throws CodeGenerationException{
    SymbolID id = ext.getSymbolID();
    String epv = IOR.getEPVVar(IOR.PUBLIC_EPV);
    final Method castMethod = 
      IOR.getBuiltinMethod(IOR.CAST, id);
    d_writer.writeComment("Cast method for interface and class type "
      + "conversions.", false);
    d_writer.println(C.getSymbolObjectPtr(id));
    d_writer.println(C.getFullMethodName(id, "_rmicast") + "(");
    d_writer.tab();
    d_writer.println("void* obj,");
    d_writer.println("sidl_BaseInterface* _ex)");
    d_writer.backTab();
    d_writer.println("{");
    d_writer.tab();
    d_writer.println(C.getSymbolObjectPtr(id) + " cast = " + C.NULL + ";");
    d_writer.println();

    d_writer.println("if(!connect_loaded) {");
    d_writer.tab();
    d_writer.println("sidl_rmi_ConnectRegistry_registerConnect(\""+id.getFullName()+
                     "\", (void*)"+IOR.getSymbolName(id)+"__IHConnect, _ex);");
    d_writer.println("connect_loaded = 1;");
    d_writer.backTab();
    d_writer.println("}");


    d_writer.println("if (obj != " + C.NULL + ") {");
    d_writer.tab();
    d_writer.println("struct sidl_BaseInterface__object* base = "
                     +"(struct sidl_BaseInterface__object*) obj;");
    d_writer.println("cast = (" + C.getSymbolObjectPtr(id) + ") (*base->" + epv 
                     + "->" + IOR.getVectorEntry(castMethod.getLongMethodName()) + ")(");
    d_writer.tab();
    d_writer.println("base->d_object,");
    d_writer.println("\"" + id.getFullName() + "\", _ex); SIDL_CHECK(*_ex);");
    d_writer.backTab();
    d_writer.backTab();
    d_writer.println("}");
    d_writer.println();
    d_writer.println("return cast;");
    d_writer.println("EXIT:");
    d_writer.println("return NULL;");
    d_writer.backTab();
    d_writer.println("}");
    d_writer.println();

  }

  public void generateConnectInternal(Extendable ext) {
    SymbolID id = ext.getSymbolID();
    
    d_writer.writeComment("RMI connector function for the class.", false);
    d_writer.println(C.getSymbolObjectPtr(id));
    d_writer.println(C.getFullMethodName(id, "_connectI") 
                     + "(const char* url, sidl_bool ar, " +
                     IOR.getExceptionFundamentalType() +
                     "*_ex)");
      
    d_writer.println("{");
    d_writer.tab();
    
    d_writer.println("return " + C.getFullMethodName(id,"_remoteConnect") 
                     + "(url, ar, _ex);");
    
    d_writer.backTab();
    d_writer.println("}");
    d_writer.println(); 

  }
}
