//
// File: ParseTree2ASTVisitor
// Package:
// Copyright: (c) 2005 The Regents of the University California
// Release: $Name$
// Revision: @(#) $Id$
// Description:

package gov.llnl.babel.parsers.sidl2;

import gov.llnl.babel.ast.SIDLFile;
import gov.llnl.babel.ast.Attribute;
import gov.llnl.babel.ast.AttributeList;
import gov.llnl.babel.ast.ImportClause;
import gov.llnl.babel.ast.RequireClause;
import gov.llnl.babel.ast.Package;
import gov.llnl.babel.ast.Version;
import gov.llnl.babel.ast.Name;
import gov.llnl.babel.ast.ScopedID;
import gov.llnl.babel.ast.TypeSearchClause;
import gov.llnl.babel.ast.Enumeration;
import gov.llnl.babel.ast.Extendable;
import gov.llnl.babel.ast.EnumItem;
import gov.llnl.babel.ast.INameable;
import gov.llnl.babel.ast.ASTNode;
import gov.llnl.babel.ast.IntLiteral;
import gov.llnl.babel.msg.MsgList;
import gov.llnl.babel.msg.UserMsg;
import gov.llnl.babel.ast.InterfaceType;
import gov.llnl.babel.ast.IAttributable;
import gov.llnl.babel.ast.ClassType;
import gov.llnl.babel.ast.NamedType;
import gov.llnl.babel.ast.ImplementsList;
import gov.llnl.babel.ast.Method;
import gov.llnl.babel.ast.MethodList;
import gov.llnl.babel.ast.Type;
import gov.llnl.babel.ast.FixedType;
import gov.llnl.babel.ast.ArrayType;
import gov.llnl.babel.ast.RArrayType;
import gov.llnl.babel.ast.Extents;
import gov.llnl.babel.ast.Argument;
import gov.llnl.babel.ast.ArgumentList;
import gov.llnl.babel.ast.ThrowsList;
import gov.llnl.babel.ast.MethodName;
import gov.llnl.babel.ast.FromClause;
import gov.llnl.babel.ast.StructType;
import gov.llnl.babel.ast.StructItem;


import java.util.Stack;
import java.util.LinkedList;
import java.util.Iterator;

/**
 * This class turns the JJTree/JavaCC generated parse tree into a bona fide AST.
 * The output AST can then be traversed for syntax checking, cloning,
 * modification, and (eventually) better code generators.
 */
public class ParseTree2ASTVisitor implements SIDLParserVisitor {

  protected Stack d_prefixStack; // items are pushed on from above, typically popped in same context
                                 // children often peek, but never pop
  
  protected Stack d_postfixStack; // items are pushed on by children as return values to parents

  protected LinkedList d_attrs;
  
  protected MsgList d_msgs;

  protected void checkAttrs( String typeName, IAttributable node, Object data, String legal_attrs[] ) { 
    if (data != null && data instanceof AttributeList ) { 
      AttributeList attrs = (AttributeList) data;  
      for (Iterator it = attrs.iterator(); it.hasNext(); ) {
        Attribute a = (Attribute) it.next();
        boolean allowable = false;
        a.setParent( (ASTNode) node); 
        if ( a.isBuiltin() ) { 
          for( int i=0; i<legal_attrs.length; i++) { 
            if ( a.getVisibleKey().equals(legal_attrs[i]) ) { 
              allowable = true;
              break;
            }
          }
        } else { 
          allowable = true;
        }
        if (allowable) {
          node.getAttributeList().addAttribute(a);
        } else {
          d_msgs.addMsg( new UserMsg( UserMsg.WARNING, 
              typeName +" cannot be " + a.getVisibleKey() + ", ignoring attribute.", a ));
        }
      }
    }
  }
  
  private boolean checkVersionedPackage(String typeName, NamedType t ) {
    if (t.getParent() instanceof Package) { 
      Package parent = (Package) t.getParent(); 
      if ( parent.getVersion() == null ) {  
        d_msgs.addMsg(new UserMsg(UserMsg.ERROR, typeName + 
              " must appear inside a versioned package.", t ));         
        return false;
      }
    } else { 
      d_msgs.addMsg(new UserMsg(UserMsg.ERROR, typeName + " must appear inside a package.", t));
      return false;
    }
    return true;
  }

  public ParseTree2ASTVisitor(MsgList msgs) {
    d_prefixStack = new Stack();
    d_postfixStack = new Stack();
    d_msgs = msgs;
  }

  public Object visit(ParseTreeNode src, Object data) {
    Object retval = null;
    switch (src.getID()) {

    case SIDLParserTreeConstants.JJTSTART:
      {
        SIDLFile f = new SIDLFile(src, src.name);
        LinkedList attrs = new LinkedList();
        d_prefixStack.push(f);
        src.childrenAccept(this, attrs);
        retval = d_prefixStack.pop();
        while( d_postfixStack.size() > 0) {
          System.err.println("Internal Error:  postfixStack has lingering item(s)" + d_postfixStack.pop());
        }
      }
      break;

    case SIDLParserTreeConstants.JJTREQUIRE:
      {
        SIDLFile parent = (SIDLFile) d_prefixStack.peek();
        RequireClause c = new RequireClause(src, parent, null, null);
        parent.appendTypeSearchClause(c);
        d_prefixStack.push(c);
        retval = src.childrenAccept(this, data);
        d_prefixStack.pop();
        if ( d_postfixStack.peek() instanceof ScopedID ) { 
          c.setScopedID((ScopedID) d_postfixStack.pop());
        }
      }
      break;

    case SIDLParserTreeConstants.JJTIMPORT:
      {
        SIDLFile parent = (SIDLFile) d_prefixStack.peek();
        ImportClause c = new ImportClause(src, parent, null, null);
        parent.appendTypeSearchClause(c);
        d_prefixStack.push(c);
        retval = src.childrenAccept(this, data);
        d_prefixStack.pop();
        if ( d_postfixStack.peek() instanceof ScopedID ) { 
          c.setScopedID((ScopedID) d_postfixStack.pop());
        }
      }
      break;

    case SIDLParserTreeConstants.JJTNAME:
      {
        INameable t = (INameable) d_prefixStack.peek();
        t.setName(new Name(src, (ASTNode) t));
        src.childrenAccept(this, data);
      }
      break;
      
    case SIDLParserTreeConstants.JJTPACKAGE:
      {
        Package p;
        Object obj = d_prefixStack.peek();
        if (obj instanceof Package) {
          Package parent = (Package) obj;
          p = new Package(src, parent, null, null);
          parent.appendNamedType(p);
        } else { // if ( obj instanceof SIDLFile ) {
          SIDLFile file = (SIDLFile) obj;
          p = new Package(src, file, null, null);
          file.appendPackage(p);
        }
        String legal_attrs[] = {"final"};
        AttributeList a = (AttributeList) data;
        a.setParent(p);
        checkAttrs( "Packages", p, a, legal_attrs);
        d_prefixStack.push(p);
        retval = src.childrenAccept(this, data);
        d_prefixStack.pop();
      }
      break;

    case SIDLParserTreeConstants.JJTSCOPEDID:
      {
        Object obj = d_prefixStack.peek();
        if (obj instanceof Package) {
          Package p = (Package) obj;
          p.setName(new Name(src));
        } else if (obj instanceof Extendable ) {
          LinkedList list = (LinkedList) data; // extends list or implements list
          ScopedID id = new ScopedID(src, (ASTNode) obj);
          list.add(id);
        } else if (obj instanceof Method ) { 
          Method m = (Method) obj;
          m.setReturnType(new ScopedID(src,m));
        } else if (obj instanceof ArrayType ) { 
          ArrayType a = (ArrayType) obj;
          a.setScalarType(new ScopedID(src,a));
        } else if (obj instanceof ThrowsList ) { 
          ThrowsList tl = (ThrowsList) obj;
          tl.addException(new ScopedID(src,tl));
        } else if (obj instanceof FromClause) { 
          FromClause f = (FromClause) obj;
          f.setScopedID(new ScopedID(src,f));
        } else { 
          d_postfixStack.push( new ScopedID(src, (ASTNode) obj) );
        }
      }
      break;

    case SIDLParserTreeConstants.JJTVERSION:
      {
        Version v = new Version(src.name);
        Object obj = d_prefixStack.peek();
        if (obj instanceof Package) {
          Package p = (Package) obj;
          p.setVersion(v);
        } else if (obj instanceof TypeSearchClause) {
          TypeSearchClause ts = (TypeSearchClause) obj;
          ts.setVersion(v);
        }
      }
      break;

    case SIDLParserTreeConstants.JJTUSERTYPE:
      {
        AttributeList attrs = new AttributeList(null,null);
        retval = src.childrenAccept(this, attrs);
      }
      break;

    case SIDLParserTreeConstants.JJTTYPEATTRS:
      retval = src.childrenAccept(this, data);
      break;

    case SIDLParserTreeConstants.JJTTYPEATTR:
    case SIDLParserTreeConstants.JJTCUSTOMATTR:
      {
        AttributeList attrs = (AttributeList) data;
        attrs.addAttribute(new Attribute(src));
        retval = src.childrenAccept(this, data);
      }
      break;

    case SIDLParserTreeConstants.JJTENUM:
      {
        ASTNode parent = (ASTNode) d_prefixStack.peek();
        Enumeration e = new Enumeration(src, parent, null);
        if ( checkVersionedPackage("Enum", e) ) { 
          // String legal_attrs[] = {};
          // AttributeList a = (AttributeList) data;
          // a.setParent(e);
          // checkAttrs( "Enums", e, a, legal_attrs);
          d_prefixStack.push(e);
          retval = src.childrenAccept(this, data);
          d_prefixStack.pop();
          ((Package) parent).appendNamedType(e);
        }
      }
      break;

    case SIDLParserTreeConstants.JJTENUMERATOR:
      {
        Enumeration e = (Enumeration) d_prefixStack.peek();
        EnumItem i = new EnumItem(src, e);
        e.addEnumItem(i);
        d_prefixStack.push(i);
        retval = src.childrenAccept(this, data);
        d_prefixStack.pop();
      }
      break;


    case SIDLParserTreeConstants.JJTSTRUCT:
      { 
        ASTNode parent = (ASTNode) d_prefixStack.peek();
        StructType s = new StructType(src, parent, null);
        if ( checkVersionedPackage("Struct", s) ) { 
          // String legal_attrs[] = {};
          // AttributeList a = (AttributeList) data;
          // a.setParent(e);
          // checkAttrs( "Enums", e, a, legal_attrs);
          d_prefixStack.push(s);
          retval = src.childrenAccept(this, data);
          d_prefixStack.pop();
          ((Package) parent).appendNamedType(s);
        }
      }
      break;
      
    case SIDLParserTreeConstants.JJTSTRUCTITEM:
      {
        StructType s = (StructType) d_prefixStack.peek();
        StructItem i = new StructItem(src, s);
        s.addStructItem(i);
        d_prefixStack.push(i);
        retval = src.childrenAccept(this, data);
        d_prefixStack.pop();
        if (d_postfixStack.size() > 0 ) { 
          if (d_postfixStack.peek() instanceof Type ) { 
            Type t = (Type) d_postfixStack.pop();
            i.setType(t);
            t.setParent(i); 
          }
        }
      }
      break;
      
    case SIDLParserTreeConstants.JJTINTEGER:
      {
        ASTNode obj = (ASTNode) d_prefixStack.peek();
        IntLiteral ilit = new IntLiteral( src, obj );
        if (obj instanceof EnumItem) {
          EnumItem e = (EnumItem) obj;
          e.setExplicitValue(ilit);
          if (ilit.isLong()) { 
            d_msgs.addMsg( new UserMsg( UserMsg.WARNING, 
                  "Long int literals not supported in enums, dropping L|l suffix.", e.getExplicitValue()));  
          }
        }
      }
      break;

    case SIDLParserTreeConstants.JJTINTERFACE:
      { 
        ASTNode parent = (ASTNode) d_prefixStack.peek();
        InterfaceType iface = new InterfaceType( src, parent, null);
        if ( checkVersionedPackage("Interface", iface) ) { 
          ((Package) parent).appendNamedType(iface);
          AttributeList a = (AttributeList) data;
          a.setParent(iface);
          String legal_attrs[] = {};
          checkAttrs("Interfaces", iface, a, legal_attrs);
          d_prefixStack.push(iface);
          retval = src.childrenAccept(this, data);
          d_prefixStack.pop();
        }
      }
      break;

    case SIDLParserTreeConstants.JJTCLASS:
    { 
      ASTNode parent = (ASTNode) d_prefixStack.peek();

      ClassType cls = new ClassType( src, parent, null);
      if ( checkVersionedPackage("Class", cls) ) { 
        ((Package) parent).appendNamedType(cls);
        AttributeList a = (AttributeList) data;
        a.setParent(cls);
        String legal_attrs[] = {"abstract"};
        checkAttrs("Classes", cls, data, legal_attrs);
        d_prefixStack.push(cls);
        retval = src.childrenAccept(this, data);
        d_prefixStack.pop();
      }
    }
    break;
       
    case SIDLParserTreeConstants.JJTEXTENDSLIST:
      { 
        InterfaceType iface = (InterfaceType) d_prefixStack.peek();
        LinkedList extendsList = iface.getExtends();
        retval = src.childrenAccept(this, extendsList);
      }
      break;
      

    case SIDLParserTreeConstants.JJTEXTENDSONE:
      { 
        LinkedList extendsOne = new LinkedList();
        retval = src.childrenAccept(this, extendsOne);
        ClassType cls = (ClassType) d_prefixStack.peek();
        Iterator it = extendsOne.iterator();
        cls.setExtends( (ScopedID) it.next() );
        while ( it.hasNext() ) {
          d_msgs.addMsg(new UserMsg(UserMsg.ERROR,"Classes can extend at most one class", (ScopedID) it.next()));
        }
      }
      break;
      
    case SIDLParserTreeConstants.JJTIMPLEMENTSLIST:
      { 
        ClassType parent = (ClassType) d_prefixStack.peek();
        if ( parent.getImplementsList() == null ) { 
          ImplementsList implementsList = new ImplementsList(src,parent);
          parent.setImplementsList( implementsList );
        }
        retval = src.childrenAccept(this, parent.getImplementsList());
      }
      break;
      
    case SIDLParserTreeConstants.JJTIMPLEMENTSALLLIST:
      {
        ClassType parent = (ClassType) d_prefixStack.peek();
        if ( parent.getImplementsList() == null ) { 
          ImplementsList implementsList = new ImplementsList(src,parent);
          parent.setImplementsList( implementsList );
        }
        ImplementsList list2 = new ImplementsList(src,parent);
        retval = src.childrenAccept(this, list2);
        for(Iterator it = list2.iterator(); it.hasNext(); ) { 
          parent.getImplementsList().addImplementsAll( (ScopedID) it.next());
       }
      }
      break;
      
    case SIDLParserTreeConstants.JJTMETHOD:
      {
        Extendable t = (Extendable) d_prefixStack.peek();
        MethodList parent = t.getMethodList();
        Method m = new Method(src,parent);
        parent.addMethod(m);
        AttributeList attrs = m.getAttributeList();
        attrs.setParent(m);
        d_prefixStack.push(m);
        retval = src.childrenAccept(this, attrs);
        d_prefixStack.pop();
      }
      break;
      
    case SIDLParserTreeConstants.JJTMETHODNAME:
      {
        ASTNode node = (ASTNode) d_prefixStack.peek();
        MethodName name = new MethodName(src,node);
        if ( node instanceof Method ) {
          Method m = (Method) node;
          m.setMethodName(name);
        } else if (node instanceof FromClause ) { 
          FromClause f = (FromClause) node;
          f.setMethodName(name);
        }
        d_prefixStack.push(name);
        retval = src.childrenAccept(this,data);
        d_prefixStack.pop();
      }
      break;
      
    case SIDLParserTreeConstants.JJTMETHODATTRS:
      retval = src.childrenAccept(this, data);
      break;

    case SIDLParserTreeConstants.JJTMETHODATTR:
      {
        AttributeList attrs = (AttributeList) data;
        attrs.addAttribute(new Attribute(src));
        retval = src.childrenAccept(this, data);
      }
      break;
      
    case SIDLParserTreeConstants.JJTRETURNTYPE:
      {
        src.childrenAccept(this, data); 
        if (d_postfixStack.size() > 0 ) { 
          if (d_postfixStack.peek() instanceof Type ) { 
            Type t = (Type) d_postfixStack.pop();
            Method m = (Method) d_prefixStack.peek();
            m.setReturnType(t);
            t.setParent(m); 
          }
        }
      }
      break;

    case SIDLParserTreeConstants.JJTSHORTNAME:
      { 
        MethodName m = (MethodName) d_prefixStack.peek();
        m.setShortName(new Name(src,m));
      }
      break;   
      
    case SIDLParserTreeConstants.JJTEXTENSION:
      {
        MethodName m = (MethodName) d_prefixStack.peek();
        m.setNameExtension(new Name(src, m));
      }
      break;
      
    case SIDLParserTreeConstants.JJTARGLIST:
      {
        retval = src.childrenAccept(this,data);
      }
      break;
      
    case SIDLParserTreeConstants.JJTARG:
      {
        Method m = (Method) d_prefixStack.peek();
        ArgumentList al = m.getArgumentList();
        Argument arg = new Argument(src,al);
        al.addArgument(arg);
        d_prefixStack.push(arg);
        retval = src.childrenAccept(this,data);
        d_prefixStack.pop();
        if (d_postfixStack.peek() instanceof Type ) { 
          arg.setType( (Type) d_postfixStack.pop() );
        }
      }
      break;
    
    case SIDLParserTreeConstants.JJTARGATTRS:
      {
        AttributeList attrs = new AttributeList(null,(Argument)d_prefixStack.peek());
        retval = src.childrenAccept(this, attrs);
      }
      break;
      
    case SIDLParserTreeConstants.JJTARGATTR:
      break;
      
    case SIDLParserTreeConstants.JJTMODE:
      {
        Argument arg = (Argument) d_prefixStack.peek();
        arg.setMode(src.name);
      }
      break;

    case SIDLParserTreeConstants.JJTARRAY:
      { 
        ASTNode parent = (ASTNode) d_prefixStack.peek();
        ArrayType a = new ArrayType(src,parent);
        d_prefixStack.push(a);
        retval = src.childrenAccept(this, data);
        d_prefixStack.pop();
        
        if (d_postfixStack.peek() instanceof FixedType ) { 
          a.setScalarType( (FixedType) d_postfixStack.pop() );
        }
        d_postfixStack.push(a);
      }
      break;
      
    case SIDLParserTreeConstants.JJTRARRAY:
    { 
      Argument parent = (Argument) d_prefixStack.peek();
      if ( parent.getMode() == "out") { 
        d_msgs.addMsg(new UserMsg(UserMsg.ERROR,"Raw Arrays cannot be \"out\" parameters", parent ));
      }
      RArrayType a = new RArrayType(src,parent);
      d_prefixStack.push(a);
      retval = src.childrenAccept(this, data);
      d_prefixStack.pop();
      
      if (d_postfixStack.peek() instanceof FixedType ) { 
        a.setScalarType( (FixedType) d_postfixStack.pop() );
      }
      d_postfixStack.push(a);
    }
    break;
      
    case SIDLParserTreeConstants.JJTSCALARTYPE:
      { 
        retval=src.childrenAccept(this,data);
      }
      break;
      

    case SIDLParserTreeConstants.JJTDIMENSION:
      {
        ASTNode obj = (ASTNode) d_prefixStack.peek();
        IntLiteral ilit = new IntLiteral( src, obj );
        if ( obj instanceof ArrayType ) { 
          ArrayType t = (ArrayType) obj;
          if ( ilit.isLong()) { 
            d_msgs.addMsg(new UserMsg(UserMsg.WARNING,
                    "Long int literals are not supported in array dimension", ilit));            
          }
          int dim = ilit.getInt();
          if ( dim < 1 || dim > 7) { 
            d_msgs.addMsg(new UserMsg(UserMsg.ERROR,
                "Array dimension must be between 1 and 8 (inclusive)", ilit));
          }
          t.setDimension( ilit.getInt() );
        }
      }
      break;
      
    case SIDLParserTreeConstants.JJTORIENTATION:
      {
        ArrayType a = (ArrayType) d_prefixStack.peek();
        if (src.name == "row-major") { 
          a.setOrientation('r');
        } else if (src.name == "column-major") { 
          a.setOrientation('c');
        } else { 
        }
      }
      break;
      

    case SIDLParserTreeConstants.JJTEXTENTS:
      {
        RArrayType a = (RArrayType) d_prefixStack.peek();
        Extents ext = new Extents(src,a);
        a.setExtents(ext);
        d_prefixStack.push(ext);
        retval = src.childrenAccept(this,data);
        d_prefixStack.pop();
      }
      break;
      
    case SIDLParserTreeConstants.JJTEXTENT:
      break;
      
    case SIDLParserTreeConstants.JJTPRIMATIVETYPE:
      {
        ASTNode parent = (ASTNode) d_prefixStack.peek();
        d_postfixStack.push(new FixedType( src, parent, src.name ));
      }
      break;


    case SIDLParserTreeConstants.JJTEXCEPTCLAUSE:
      {
        Method m = (Method) d_prefixStack.peek();
        ThrowsList tl = m.getThrowsList();
        d_prefixStack.push(tl);
        retval = src.childrenAccept(this,data);
        d_prefixStack.pop();
      }
      break;
    

    case SIDLParserTreeConstants.JJTFROMCLAUSE:
      { 
        Method m = (Method) d_prefixStack.peek();
        FromClause f = new FromClause(src,m);
        m.setFromClause(f);
        d_prefixStack.push(f);
        retval = src.childrenAccept(this, data);
        d_prefixStack.pop();
      }
      break;
      
    case SIDLParserTreeConstants.JJTNUMBER:
    case SIDLParserTreeConstants.JJTMULT:
    case SIDLParserTreeConstants.JJTADD:
    case SIDLParserTreeConstants.JJTASSERTION:
    case SIDLParserTreeConstants.JJTENSURES:
    case SIDLParserTreeConstants.JJTSHIFT:
    case SIDLParserTreeConstants.JJTGTLT:
    case SIDLParserTreeConstants.JJTAND:
    case SIDLParserTreeConstants.JJTOR:
    case SIDLParserTreeConstants.JJTEQUALITY:
    case SIDLParserTreeConstants.JJTCOMPLEX:
    case SIDLParserTreeConstants.JJTLITERAL:
    case SIDLParserTreeConstants.JJTFUNCARGS:
    case SIDLParserTreeConstants.JJTUNARY:
    case SIDLParserTreeConstants.JJTFUNC:
    case SIDLParserTreeConstants.JJTPOWER:
    case SIDLParserTreeConstants.JJTIMPLICATION:
      retval = src.childrenAccept(this, data);
      break;
    default:
      System.out.println(src + ": acceptor unimplemented in subclass?");
    }
    return retval;
  }
}
