package gov.llnl.babel.visitor;

import gov.llnl.babel.ast.ASTNode;
import gov.llnl.babel.ast.Name;
import gov.llnl.babel.ast.Package;
import gov.llnl.babel.ast.SIDLFile;
import gov.llnl.babel.msg.MsgList;
import gov.llnl.babel.msg.UserMsg;
import gov.llnl.babel.msg.UserMsg2;
import gov.llnl.babel.ast.Attribute;
import gov.llnl.babel.ast.AttributeList;
import gov.llnl.babel.ast.Enumeration;
import gov.llnl.babel.ast.EnumItem;
import gov.llnl.babel.ast.MethodList;
import gov.llnl.babel.ast.Method;
import gov.llnl.babel.ast.ClassType;
import gov.llnl.babel.ast.InterfaceType;
import gov.llnl.babel.ast.Argument;
import gov.llnl.babel.ast.ArgumentList;
import gov.llnl.babel.ast.StructType;
import gov.llnl.babel.ast.StructItem;
import gov.llnl.babel.ast.RArrayType;
import gov.llnl.babel.ast.Type;
import gov.llnl.babel.ast.FixedType;

import java.util.Iterator;
import java.util.HashMap;
import java.util.TreeSet;
import java.util.LinkedList;

/**
 * Checks for naming collisions two args with same name in same method, 
 * two Types with same name in a packgage, and so on for all names in 
 * the ast.  For enumerations, it also checks for duplicate explicit values
 * and assigns unique values to items otherwise unassigned.  For packages, it
 * also blocks attempts to reopen a final package.  Naturally, it will detect
 * collisions for packages that are closed and reopened.
 * 
 * @author kumfert
 *
 */
public class CollisionChecker extends Visitor {

  protected MsgList d_msgs;
  
  private boolean d_hasRArrays;
  
  class Tuple { 
    ASTNode node;
    HashMap members;
    public Tuple( ASTNode n, HashMap m) { 
      node = n; 
      members = m;
    }
  }
  
  static protected HashMap d_symbolTable = new HashMap(); 

  protected HashMap d_currentScope;
  
	public CollisionChecker( MsgList msgs ) {
		super();
    d_msgs = msgs;
    d_currentScope = d_symbolTable;
	}
  
  protected AttributeList checkAttributes( AttributeList attrs ) { 
    if ( attrs == null ) { return null; }
    AttributeList l = (AttributeList) attrs.cloneEmpty();
    HashMap m = new HashMap();
    for ( Iterator i = attrs.iterator(); i.hasNext(); ) { 
      Attribute a = (Attribute) i.next();
      String str_name = a.getKey();
      if ( m.containsKey(str_name) ) {
        Attribute b = (Attribute) m.get( str_name );
        if (a.isBuiltin()) { 
          if ( a.getValue().equals( b.getValue() )) { 
            d_msgs.addMsg( new UserMsg2( UserMsg.WARNING, "redundant built-in attribute \"" + a + "\"", a, 
                "is already declared here", b )); 
          } else { 
            d_msgs.addMsg( new UserMsg2( UserMsg.ERROR, "conflicting built-in attribute \"" + a  + "\"", a, 
                "cannot be used in conjunction with preceding \"" + b + "\" attribute", b )); 
          }
        } else { 
          if ( a.getValue().equals( b.getValue() )) { 
            d_msgs.addMsg( new UserMsg2( UserMsg.WARNING, "redundant attribute \"" + a + "\"", a, 
                "is already declared here", b )); 
          } else { 
            d_msgs.addMsg( new UserMsg2( UserMsg.ERROR, "conflicting attribute \"" + a.getValue() + "\"", a, 
                "cannot be used after \"" + b + "\"", b )); 
          }
        }
      } else { 
        m.put(str_name, a);
        l.addAttribute(a);
      }    
    }
    return l;
  }
  
  protected void checkName( String scope_name, Name name, HashMap scope ) { 
    String str_name = name.toString();
    if ( scope.containsKey(str_name) ) {
      d_msgs.addMsg( new UserMsg2( UserMsg.ERROR, "name collision on \"" + str_name + "\"", name, 
          "previously used here in same " + scope_name , (ASTNode) scope.get( str_name )));
    } else { 
      scope.put(str_name, name);
    }
  }

  protected void checkNameInPackageScope( Name name, HashMap scope ) { 
    String str_name = name.toString();
    if ( scope.containsKey(str_name) ) {
      if ( scope.get(str_name) instanceof Tuple ) {
        Tuple t = (Tuple) scope.get(str_name);
        d_msgs.addMsg( new UserMsg2( UserMsg.ERROR, "name collision on \"" + str_name + "\"", name, 
            "previously used here in same package", t.node));
      } else { 
        Name n = (Name) scope.get(str_name);
        d_msgs.addMsg( new UserMsg2( UserMsg.ERROR, "name collision on \"" + str_name + "\"", name, 
            "previously used here in same package", n ));
      }
    } else { 
      scope.put(str_name, name);
    }
  }
  
	public Object visitSIDLFile(SIDLFile node, Object data) {
    super.visitSIDLFile(node, d_symbolTable);
		return data;
	}
	
	public Object visitPackage(Package node, Object data) {
	  node.setAttributeList( checkAttributes( node.getAttributeList() ) );
	  String str_name = node.getName().toString();
    HashMap parent_scope = (HashMap) data;
    boolean collision_allowed = false;
    if ( parent_scope.containsKey(str_name) ) { 
      if ( parent_scope.get(str_name) instanceof Tuple ) { 
        ASTNode n = ((Tuple) parent_scope.get(str_name)).node;
        if ( n instanceof Package ) { 
          Package p = (Package) n;
          if (node.getAttributeList().hasAttribute("final")) { 
            d_msgs.addMsg( new UserMsg2( UserMsg.ERROR, "final package \"" + str_name + "\"", node.getAttributeList().getAttribute("final"), 
                "previously declared reentrant", p.getName() ));
            return data;
          } else if ( p.getAttributeList().hasAttribute("final")) { 
            d_msgs.addMsg( new UserMsg2( UserMsg.ERROR, "final package \"" + str_name + "\"", p.getAttributeList().getAttribute("final"), 
                "cannot be reopened here", node.getName() ));
            return data;
          } else { 
            collision_allowed = true;
          }
        }
      }
    }
    HashMap my_scope;
    if ( collision_allowed ) { // entering reentrant package
      my_scope = ((Tuple) parent_scope.get(str_name)).members;
    } else { // do the normal thing
      checkNameInPackageScope( node.getName(), parent_scope ); 
      my_scope = new HashMap();
      parent_scope.put(str_name, new Tuple(node,my_scope)); //record my scope for later
    }
    super.visitPackage(node, my_scope);  
    return data;
	}
	
	public Object visitEnumeration(Enumeration node, Object data) {
	  checkNameInPackageScope( node.getName(), (HashMap) data );
	  HashMap new_scope = new HashMap();
    super.visitEnumeration(node, new_scope);
    // Now create a suitable list of integers for unassigned enums.
    TreeSet s = new TreeSet(); 
    int listsize = node.getEnumItemList().size();
    for( int i=0; i<listsize; i++) { 
      s.add(new Integer(i));
    }
    // Now run through list of EnumItems, checking their explicit values
    // report errors on explicit duplicates, and remove any from set of 
    // candidate numbers 
    HashMap dups = new HashMap(); // will detect duplicates
    for( Iterator i = node.getEnumItemList().iterator(); i.hasNext(); ) {
      EnumItem e = (EnumItem) i.next();
      if ( e.hasValue() ) { 
        Integer value = new Integer( e.getValue() );
        s.remove(value); 
        if (dups.containsKey(value)) { 
          EnumItem duplicate = (EnumItem) dups.get(value);
          if ( e.getExplicitValue() == null || duplicate.getExplicitValue()==null ) { 
            d_msgs.addMsg(new UserMsg2( UserMsg.ERROR, "(internal) duplicate enum values",e,
                "previously appeared(generated?!?) here", duplicate));
          } else { 
            d_msgs.addMsg( new UserMsg2( UserMsg.ERROR, "duplicate explicit enum value", e.getExplicitValue(), 
              "previously appeared here", duplicate.getExplicitValue() ));
          }
        } else { 
          dups.put(value, e);
        }
      }   
    }
    // Now we simply march through the list of remaining numbers and 
    // assign values to unassigned enum items
    Iterator sorted_remaining_ints = s.iterator();
    for( Iterator i = node.getEnumItemList().iterator(); i.hasNext(); ) {
      EnumItem e = (EnumItem) i.next();
      if ( !e.hasValue() ) { 
        Integer smallest_valid_int = (Integer) sorted_remaining_ints.next();
        e.setValue( smallest_valid_int.intValue());
      }
    }
    return data;
  }
  
  public Object visitEnumItem(EnumItem node, Object data) { 
    checkName( "enumeration", node.getName(), (HashMap) data );
    return data;
  }
  
  public Object visitStructType(StructType node, Object data) { 
    checkNameInPackageScope( node.getName(), (HashMap) data );
    HashMap new_scope = new HashMap();
    return super.visitStructType(node, new_scope);
  }
  
  public Object visitStructItem(StructItem node, Object data) { 
    checkName( "struct", node.getName(), (HashMap) data );
    return data;
  }
  
  public Object visitMethodList(MethodList node, Object data) {
    HashMap new_scope = new HashMap();
    super.visitMethodList(node, new_scope);
    return data;
  }
  
  public Object visitMethod(Method node, Object data ) { 
    ASTNode parent = node.getParent().getParent();
    String scopeName = "scope";
    if ( parent instanceof ClassType ) { 
      scopeName = "class";
    } else if ( parent instanceof InterfaceType ) { 
      scopeName = "interface";
    } 
    checkName( scopeName, node.getMethodName(), (HashMap) data );
    super.visitMethod(node,data);
    return data;
  }
  
  public Object visitArgumentList(ArgumentList node, Object data) { 
    HashMap new_scope = new HashMap();
    d_hasRArrays=false;
    super.visitArgumentList(node, new_scope);
    if (d_hasRArrays) { 
      // 1. gather up all the names of all the extents
      LinkedList names = new LinkedList();
      HashMap args = new HashMap();
      for(Iterator it = node.iterator(); it.hasNext(); ) { 
        Argument arg = (Argument) it.next();
        if ( arg.getType() instanceof RArrayType ) { 
          RArrayType t = (RArrayType) arg.getType();
          names.addAll(t.getExtents().getNames());
        }
        args.put(arg.getName().toString(), arg);
      }
      // 2. make sure each name is found in the same scope
      for(Iterator iter = names.iterator(); iter.hasNext(); ) { 
        Name n = (Name) iter.next();
        if (! new_scope.containsKey(n.toString())) { 
          d_msgs.addMsg( new UserMsg( UserMsg.ERROR, "array extent \"" + n + "\" not found in arglist", n));
        } else { 
          Argument arg = (Argument) args.get(n.toString());
          Type t = (Type) arg.getType();
          FixedType ft = t instanceof FixedType ? (FixedType) t : null;
          if (ft == null ||  !ft.toString().equals("int") ) { 
            d_msgs.addMsg( new UserMsg2( UserMsg.ERROR, "array extent \"" + n + "\" should be of type int", n, 
                        "instead of type" + t.getTypeName(), arg.getName()));
          }
        }     
      }
    }
    return data;
  }
  
  public Object visitArgument(Argument node, Object data) { 
    checkName("argument list", node.getName(), (HashMap) data);
    super.visitArgument(node,data);
    return data;
  }
  
  public Object visitRArrayType(RArrayType node, Object data) { 
    d_hasRArrays = true;
    return data;
  }
  
}
